/**
  *  \file client/dlg-draw.cc
  *  \brief Drawing-related dialogs
  */

#include "client/dlg-draw.h"
#include "cpluslib/assert.h"
#include "ui/uiglobal.h"
#include "ui/layout.h"
#include "ui/util.h"
#include "gfx/palette.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/widgets.h"
#include "client/marks.h"

/***************************** WColorSelector ****************************/

enum {
    CS_LINES   = 3,
    CS_COLUMNS = 10
};
CTASSERT(CS_LINES * CS_COLUMNS == GDrawing::NUM_USER_COLORS);

/** Constructor. */
WColorSelector::WColorSelector()
    : user_color(1)
{ }

/** Destructor. */
WColorSelector::~WColorSelector()
{ }

/** Move cursor by specified delta. */
void
WColorSelector::moveBy(int delta)
{
    int n = user_color + delta;
    if (n <= 0)
        n += GDrawing::NUM_USER_COLORS;
    if (n > GDrawing::NUM_USER_COLORS)
        n -= GDrawing::NUM_USER_COLORS;
    setColor(n);
}

/** Get size of one color box. */
int
WColorSelector::getFrameSize() const
{
    return font_heights[FONT_TITLE];
}

/** Select a color.
    \param user_color User color index. */
void
WColorSelector::setColor(int user_color)
{
    if (user_color >= 0 && user_color <= GDrawing::NUM_USER_COLORS) {
        if (user_color != this->user_color) {
            this->user_color = user_color;
            drawWidget();
        }
    }
}

/** Get selected color. */
int
WColorSelector::getColor() const
{
    return user_color;
}

void
WColorSelector::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = GfxPoint(CS_COLUMNS*getFrameSize(), CS_LINES*getFrameSize());
}

void
WColorSelector::drawContent(GfxCanvas& can)
{
    const int fs = getFrameSize();
    for (int line = 0; line < CS_LINES; ++line) {
        for (int col = 0; col < CS_COLUMNS; ++col) {
            const int this_color = 1 + col + line*CS_COLUMNS;
            const int x = getExtent().x + col*fs;
            const int y = getExtent().y + line*fs;
            // Draw frame
            if (this_color == user_color) {
                drawFrameDown(can, GfxRect(x, y, fs, fs));
            } else {
                drawFrameUp(can, GfxRect(x, y, fs, fs));
            }

            // Draw color box
            drawSolidBar(can, GfxRect(x+1, y+1, fs-2, fs-2),
                         standard_colors[GDrawing::getColorFromUserColor(this_color)]);
        }
    }
    can.update(getExtent());
}

bool
WColorSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse && e.mouse.b != 0 && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        // Mouse clicked inside widget
        setSelectedWidget();
        int col  = (e.mouse.x - getExtent().x) / getFrameSize();
        int line = (e.mouse.y - getExtent().y) / getFrameSize();
        if (col < CS_COLUMNS && line < CS_LINES)
            setColor(1 + col + line*CS_COLUMNS);
        return true;
    }
    if (e.type == ev_Key) {
        // Handle some keys
        switch (e.key.code) {
         case SDLK_LEFT:
            setSelectedWidget();
            moveBy(-1);
            return true;
         case SDLK_RIGHT:
            setSelectedWidget();
            moveBy(+1);
            return true;
         case SDLK_UP:
            setSelectedWidget();
            moveBy(-10);
            return true;
         case SDLK_DOWN:
            setSelectedWidget();
            moveBy(+10);
            return true;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

/** Perform standard dialog.
    \param title Window title
    \param offer_adjacent true to offer "Adjacent" button
    \return rCancel, rOK, rAdjacent */
int
WColorSelector::doStandardDialog(string_t title, bool offer_adjacent)
{
    ObjectHolder h;
    UIWindow w(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    w.setLayout(UIVBoxLayout::instance5);
    w.add(h.add(new UIStaticText(_("Choose color:"), UIStaticText::BlackTitle)));
    w.add(*this);
    g.add(h.add(new UIButton(rOK, _("OK"), SDLK_RETURN)).addStop());
    if (offer_adjacent)
        g.add(h.add(new UIButton(rAdjacent, _("Adjacent"), 'a')).addStop());
    g.add(h.add(new UIButton(rCancel, _("Cancel"), SDLK_ESCAPE)).addStop());
    g.add(h.add(new UISpacer()));
    w.add(g);
    w.pack();
    ui_root->centerWidget(w);
    return ui_root->runModal(w);
}

/************************** WMarkerTypeSelector **************************/

WMarkerTypeSelector::WMarkerTypeSelector()
    : marker_type(0)
{ }

WMarkerTypeSelector::~WMarkerTypeSelector()
{ }

/** Move cursor by specified delta. */
void
WMarkerTypeSelector::moveBy(int delta)
{
    int n = marker_type + delta;
    if (n <= 0)
        n += NUM_USER_MARKERS;
    if (n > NUM_USER_MARKERS)
        n -= NUM_USER_MARKERS;
    setMarkerType(n);
}

/** Get size of one color box. */
int
WMarkerTypeSelector::getFrameSize() const
{
    return font_heights[FONT_TITLE];
}

void
WMarkerTypeSelector::setMarkerType(int marker_type)
{
    if (marker_type >= 0 && marker_type < NUM_USER_MARKERS) {
        if (marker_type != this->marker_type) {
            this->marker_type = marker_type;
            drawWidget();
        }
    }
}
int
WMarkerTypeSelector::getMarkerType() const
{
    return marker_type;
}

void
WMarkerTypeSelector::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = GfxPoint(NUM_USER_MARKERS*getFrameSize(), getFrameSize());
}

void
WMarkerTypeSelector::drawContent(GfxCanvas& can)
{
    GfxParameters& parm = getSkin();
    parm.color = standard_colors[COLOR_BLACK];

    const int fs = getFrameSize();
    for (int col = 0; col < NUM_USER_MARKERS; ++col) {
        const int x = getExtent().x + col*fs;
        const int y = getExtent().y;
        // Draw frame
        if (col == marker_type) {
            drawFrameDown(can, GfxRect(x, y, fs, fs));
        } else {
            drawFrameUp(can, GfxRect(x, y, fs, fs));
        }

        // Draw box
        drawSolidBar(can, GfxRect(x+1, y+1, fs-2, fs-2), standard_colors[COLOR_GRAY]);

        // Draw marker
        drawMarker(can, parm, getUserMarker(col, true), x + fs/2, y + fs/2);
    }
    can.update(getExtent());
}

bool
WMarkerTypeSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse && e.mouse.b != 0 && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        // Mouse clicked inside widget
        setSelectedWidget();
        int col  = (e.mouse.x - getExtent().x) / getFrameSize();
        if (col < NUM_USER_MARKERS)
            setMarkerType(col);
        return true;
    }
    if (e.type == ev_Key) {
        // Handle some keys
        switch (e.key.code) {
         case SDLK_LEFT:
            setSelectedWidget();
            moveBy(-1);
            return true;
         case SDLK_RIGHT:
            setSelectedWidget();
            moveBy(+1);
            return true;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

bool
WMarkerTypeSelector::doStandardDialog(string_t title)
{
    ObjectHolder h;
    UIWindow w(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    w.setLayout(UIVBoxLayout::instance5);
    w.add(h.add(new UIStaticText(_("Choose marker type:"), UIStaticText::BlackTitle)));
    w.add(*this);
    g.add(h.add(new UIButton(true,  _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UIButton(false, _("Cancel"), SDLK_ESCAPE)).addStop());
    g.add(h.add(new UISpacer()));
    w.add(g);
    w.pack();
    ui_root->centerWidget(w);
    return ui_root->runModal(w);
}

/*************************** WNewDrawingDialog ***************************/

WNewDrawingDialog::WNewDrawingDialog()
    : UIWindow(_("Starchart Drawing"), UIWindow::BLUE_WINDOW, COLOR_GRAY)
{
    buildDialog();
}

WNewDrawingDialog::~WNewDrawingDialog()
{ }

void
WNewDrawingDialog::buildDialog()
{
    // VBox
    //   "What do you want to draw?"
    //   Grid
    //     P | Polygon
    //     R | Rectangle
    //     C | Circle
    //     M | Marker
    //   [Tag option...]
    //   HBox
    //     ESC, H

    setLayout(UIVBoxLayout::instance5);

    UIGroup& g2 = h.add(new UIGroup(h.add(new UIGridLayout(3))));
    UIGroup& g3 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g2.add(h.add(new UIButton(LineDrawing, "P", 'p')).addStop());
    g2.add(h.add(new UIStaticText(_("Polygon: set of lines, starting here"), UIStaticText::BlackTitle)));
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(RectangleDrawing, "R", 'r')).addStop());
    g2.add(h.add(new UIStaticText(_("Rectangle"), UIStaticText::BlackTitle)));
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(CircleDrawing, "C", 'c')).addStop());
    g2.add(h.add(new UIStaticText(_("Circle centered here"), UIStaticText::BlackTitle)));
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(MarkerDrawing, "M", 'm')).addStop());
    g2.add(h.add(new UIStaticText(_("Marker"), UIStaticText::BlackTitle)));
    g2.add(h.add(new UISpacer()));

    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(-1, _("Cancel"), SDLK_ESCAPE)).addStop());

    add(h.add(new UIStaticText(_("What do you want to draw into the starcharts?"),
                               UIStaticText::BlackTitle)));
    add(g2);
    add(g3);

    pack();
}

bool
WNewDrawingDialog::runDialog()
{
    ui_root->centerWidget(*this);
    int result = ui_root->runModal(*this);
    if (result == LineDrawing || result == RectangleDrawing || result == CircleDrawing || result == MarkerDrawing) {
        type = GDrawingType(result);
        return true;
    } else {
        return false;
    }
}

void
WNewDrawingDialog::setColor(uint8 user_color)
{
    this->user_color = user_color;
}

uint8
WNewDrawingDialog::getColor() const
{
    return user_color;
}

void
WNewDrawingDialog::setTag(atom_t tag)
{
    this->tag = tag;
}

atom_t
WNewDrawingDialog::getTag() const
{
    return tag;
}

GDrawingType
WNewDrawingDialog::getType() const
{
    return type;
}
