/**
  *  \file client/scr-base.cc
  *  \brief Starbase Screen
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "client/cscreen.h"
#include "client/marks.h"
#include "game/global.h"
#include "gfx/font.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "ui/msgbox.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "client/actions.h"
#include "game/ship.h"
#include "ui/listbox.h"
#include "ui/widgets.h"
#include "client/dlg-tech.h"
#include "ui/table.h"
#include "client/dlg-structure.h"
#include "client/dlg-ammo.h"
#include "client/dlg-build.h"
#include "resmgr/resid.h"
#include "game/tables.h"
#include "client/wscanchart.h"
#include "client/dlg-comment.h"
#include "int/intglobal.h"
#include "game/config.h"
#include "game/unitscore.h"

namespace {
    // FIXME?
    enum { cm_Base_Recycle = cm_ControlScreenFirst + 2000 };

    class WBaseSelect : public WObjectSelectionTile {
     public:
        WBaseSelect(GBaseType& type);
        void drawContent(GfxCanvas& can);
    };

    class WBaseScreen : public WControlScreen {
     public:
        WBaseScreen();
        ~WBaseScreen();
        bool handleEvent(const UIEvent& e, bool second_pass);
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        void update();
    };

    class WBaseScreenHeaderTile : public WControlScreenHeaderTile {
     public:
        WBaseScreenHeaderTile(GObjectType& type);
        string_t getTitle();
        string_t getSubtitle();
        ResId getPictureId();
    };

    class WBaseMineralTile : public WObjectTypeObserverTile {
        void drawBaseMineralTile(UIAbstractTableWriter& tw);
     public:
        WBaseMineralTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };

    class WBaseTechTile : public WObjectTypeObserverTile {
        void drawBaseTechTile(UIAbstractTableWriter& tw);
        ObjectHolder h;
     public:
        WBaseTechTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };

    class WBaseOrderTile : public WObjectTypeObserverTile {
        ObjectHolder h;
     public:
        WBaseOrderTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };
}

WBaseSelect::WBaseSelect(GBaseType& type)
    : WObjectSelectionTile(type)
{ }

void
WBaseSelect::drawContent(GfxCanvas& can)
{
    // can static_cast to GBaseType
    GPlanet& pp = getType().getUniverse()->getPlanet(getCurrent());
    UISkin& sk = getSkin();

    sk.startDraw();

    int y = getExtent().y, x = getExtent().x, rem = 0;
    if (pp.isMarked()) {
        sk.drawBackground(can, GfxRect(x, y, 15, font_heights[FONT_NORMAL]));
        sk.getParameters().color = standard_colors[COLOR_GREENBLACK];
        drawSelection(can, sk, extent.x + 5, extent.y + 7, 1, 2);
        rem += 15;
    }
    sk.setColor(UIColor::tc_Static);
    sk.outTextF(can, FONT_NORMAL, x+rem, y, extent.w - rem, pp.getName());
    y += font_heights[FONT_NORMAL]+2;

    sk.outTextF(can, FONT_NORMAL, x, y, extent.w,
             format(_("%d mc; %d kt supplies"),
                    pp.getCargo(el_Money), pp.getCargo(el_Supplies)));
    y += font_heights[FONT_NORMAL];

    /* minerals */
    sk.outTextF(can, FONT_NORMAL, x, y, 50, _("Minerals:"));
    sk.getParameters().just_x = 2;
    sk.outTextF(can, FONT_NORMAL, x+130, y, 80, format("%dN", pp.getCargo(el_Neutronium)));
    sk.outTextF(can, FONT_NORMAL, x+210, y, 80, format("%dT", pp.getCargo(el_Tritanium)));
    sk.outTextF(can, FONT_NORMAL, x+290, y, 80, format("%dD", pp.getCargo(el_Duranium)));
    sk.outTextF(can, FONT_NORMAL, x+370, y, 80, format("%dM", pp.getCargo(el_Molybdenum)));
    y += font_heights[FONT_NORMAL];

    /* tech */
    sk.getParameters().just_x = 0;
    sk.outTextF(can, FONT_NORMAL, x, y, 100, _("Tech Levels:"));
    sk.getParameters().just_x = 2;
    sk.outTextF(can, FONT_NORMAL, x+130, y, 30, format("%dH", pp.getBaseTech(tl_Hull)));
    sk.outTextF(can, FONT_NORMAL, x+210, y, 80, format("%dE", pp.getBaseTech(tl_Engine)));
    sk.outTextF(can, FONT_NORMAL, x+290, y, 80, format("%dB", pp.getBaseTech(tl_Beam)));
    sk.outTextF(can, FONT_NORMAL, x+370, y, 80, format("%dT", pp.getBaseTech(tl_Torpedo)));
    y += font_heights[FONT_NORMAL]+2;

    /* are we building something? */
    sk.getParameters().just_x = 0;
    if (int sid = pp.getShipCloningHere(getDisplayedTurn().getCurrentUniverse())) {
        sk.outTextF(can, FONT_NORMAL, x, y, extent.w, format(_("Cloning %s"),
                                                             getDisplayedTurn().getShip(sid).getName()));
    } else if (int hid = pp.getBaseBuildHull()) {
        sk.outTextF(can, FONT_NORMAL, x, y, extent.w, format(_("Building a %s"), getHull(hid).getName()));
    } else {
        sk.outTextF(can, FONT_NORMAL, x, y, extent.w, string_t());
    }
    y += font_heights[FONT_NORMAL];

    /* repairing something? */
    if (pp.getBaseShipyardOrder()) {
        sk.outTextF(can, FONT_NORMAL, x, y, extent.w,
                    format(pp.getBaseShipyardOrder() == 1 ? _("Repairing %s") : _("Recycling %s"),
                           getDisplayedTurn().getShip(pp.getBaseShipyardId()).getFullName()));
    } else {
        sk.outTextF(can, FONT_NORMAL, x, y, extent.w, string_t());
    }
    y += font_heights[FONT_NORMAL]+2;

    /* administrivia */
    sk.outTextF(can, FONT_NORMAL, x, y, extent.w, format(_("Planetary FCode: %s"), pp.getFCode()));
    y += font_heights[FONT_NORMAL];
    sk.outTextF(can, FONT_NORMAL, x, y, extent.w,
                format(_("SB Defense: %d; Fighters: %d; Damage: %d%%"),
                       pp.getStructures(str_BaseDefense), pp.getCargo(el_Fighters), pp.getBaseDamage()));
    y += font_heights[FONT_NORMAL];

    can.update(getExtent());
}

int
chooseBase()
{
    int nbases = global_base_type.getCount();
    if (nbases == 0) {
        messageBox(_("You do not have any starbases. "
                     "To build starbases, use [F8] on a planet."),
                   _("Starbases"));
        return 0;
    } else if (nbases == 1) {
        return global_base_type.getCurrent();
    } else {
        return WBaseSelect(global_base_type).doStandardDialog(_("Select Starbase"));
    }
}

/**************************** WBaseMineralTile ***************************/

WBaseMineralTile::WBaseMineralTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, font_heights[FONT_NORMAL] * 4), type)
{ }

void
WBaseMineralTile::drawBaseMineralTile(UIAbstractTableWriter& tw)
{
    GPlanet* p = dynamic_cast<GPlanet*>(getObject());
    if (!p)
        return;

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Neu:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Tri:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Dur:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mol:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("999,999 kt"));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", p->getCargo(el_Neutronium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", p->getCargo(el_Tritanium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", p->getCargo(el_Duranium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", p->getCargo(el_Molybdenum)));

    tw.newLColumn(5);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Money:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Supplies:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("999,999 mc"));
    tw.writeVariableCell(UIColor::tc_Green, format("%d mc", p->getCargo(el_Money)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", p->getCargo(el_Supplies)));

    tw.finish();
}

void
WBaseMineralTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, 0);
    drawBaseMineralTile(tw);
}

/***************************** WBaseTechTile *****************************/

WBaseTechTile::WBaseTechTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 4 * font_heights[FONT_NORMAL]), type)
{
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(0,  0, 20, 20), h.add(new UIButton("T", 't')));
    addControl(GfxRect(0, 25, 20, 20), h.add(new UIButton("D", 'd')));
}

void
WBaseTechTile::drawBaseTechTile(UIAbstractTableWriter& tw)
{
    GPlanet* p = dynamic_cast<GPlanet*>(getObject());
    if (!p)
        return;

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Engines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Hulls:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Beam Weapons:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Torpedoes:"));

    tw.newRColumn(2 * font_ems[FONT_NORMAL]);
    tw.writeVariableCell(UIColor::tc_Green, itoa(p->getBaseTech(tl_Engine)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(p->getBaseTech(tl_Hull)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(p->getBaseTech(tl_Beam)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(p->getBaseTech(tl_Torpedo)));

    tw.newLColumn(5);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Defense:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Fighters:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Damage:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth(_("none")));

    tw.writeVariableCell(UIColor::tc_Green, itoa(p->getStructures(str_BaseDefense)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(p->getCargo(el_Fighters)));
    if (p->getBaseDamage()) {
        tw.writeVariableCell(UIColor::tc_Red, itoa(p->getBaseDamage()) + "%");
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
    }

    tw.finish();
}

void
WBaseTechTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, 0);
    drawBaseTechTile(tw);
}

/***************************** WBaseOrderTile ****************************/

WBaseOrderTile::WBaseOrderTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 100), type)
{
    setLayout(LeftJustified, 5, true);
    addControl(GfxRect(0,  2, 20, 20), h.add(new UIButton("B", 'b')));
    addControl(GfxRect(0, 27, 20, 20), h.add(new UIButton(cm_Base_Recycle, "R", 'r')).addCommand());
    addControl(GfxRect(0, 52, 20, 20), h.add(new UIButton("M", 'm')));
    addControl(GfxRect(0, 77, 20, 20), h.add(new UIButton("A", 'a')));
}

void
WBaseOrderTile::drawData(GfxCanvas& can)
{
    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
        /* headers */
        setColor(UIColor::tc_Static);
        outText(can, FONT_NORMAL, 35, 55, _("Mission:"));
        outText(can, FONT_NORMAL, 35, 80, _("Build Fighters/Torpedoes"));

        /* ship */
        int hull = p->getBaseBuildHull();
        int clnid = p->getShipCloningHere(getDisplayedTurn().getCurrentUniverse());
        string_t text;

        if (hull) {
            if (clnid) {
                setColor(UIColor::tc_Red);
                text = format(_("Cloning, and building %s"),
                              getHull(hull).getName());
            } else {
                setColor(UIColor::tc_Yellow);
                text = format(_("Building a %s"),
                              getHull(hull).getName());
            }
        } else {
            if (clnid) {
                setColor(UIColor::tc_Yellow);
                text = format(_("Cloning %s"),
                              getDisplayedTurn().getShip(clnid).getName());
            } else {
                setColor(UIColor::tc_Static);
                text = _("Build a new ship");
            }
        }
        outTextF(can, FONT_NORMAL, 35, 5, extent.w-35, text);

        /* fix/recycle */
        if (p->getBaseShipyardOrder()) {
            setColor(UIColor::tc_Yellow);
            outTextF(can, FONT_NORMAL, 35, 30, extent.w-35,
                     format(p->getBaseShipyardOrder() == 1 /* FIXME: we have an enum for that */
                            ? _("Repairing %s")
                            : _("Recycling %s"),
                            getDisplayedTurn().getShip(p->getBaseShipyardId()).getFullName()));
        } else {
            setColor(UIColor::tc_Static);
            outTextF(can, FONT_NORMAL, 35, 30, extent.w-35, _("Repair/recycle a starship"));
        }

        /* mission */
        const int mx = 65 + fonts[FONT_NORMAL]->getTextWidth(_("Mission:"));
        setColor(UIColor::tc_Green);
        outTextF(can, FONT_NORMAL, mx, 55, extent.w-mx, getBaseMissionName(p->getBaseMission()));
    }
}

/****************************** WBaseScreen ******************************/

WBaseScreen::WBaseScreen()
    : WControlScreen(global_base_type, WScreenId::BaseScreen)
{
    addTile(holder.add(new WBaseScreenHeaderTile(global_base_type)));
    addTile(holder.add(new WBaseMineralTile(global_base_type)), _("Minerals & Funds:"));
    addTile(holder.add(new WBaseTechTile(global_base_type)),    _("Technology & Defense:"));
    addTile(holder.add(new WBaseOrderTile(global_base_type)),   _("Orders:"));
    addTile(holder.add(new WPlanetFCodeTile(global_base_type)));
    addTile(holder.add(new WPlanetBaseTile(global_base_type, true)));
    addChart(holder.add(new WScannerChartWidget()));
}

WBaseScreen::~WBaseScreen()
{ }

static void
doBaseXchg(WControlScreen& scr)
{
    WScanner& scanner = scr.getScanner();
    int pid = scanner.getScannedPlanetId(scanner.getScannedShipId() != 0);
    if (pid && pid != scr.getType().getCurrent()) {
        if (scr.getType().isExistant(pid))
            scr.stopWithScreenId(WScreenId(WScreenId::BaseScreen, pid));
        else if (global_planet_type.isExistant(pid))
            scr.stopWithScreenId(WScreenId(WScreenId::PlanetScreen, pid));
        else
            putKey(ss_Ctrl + SDLK_F1);
    } else {
        putKey(ss_Ctrl + SDLK_F1);
    }
}

bool
WBaseScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_Base_Recycle) {
        int pid         = getType().getCurrent();
        GUniverse& univ = *getType().getUniverse();
        GPlanet& me     = univ.getPlanet(pid);
        doBaseFixRecycle(univ, me, sender);
        return true;
    }
    return WControlScreen::handleCommand(cmd, sender);
}

bool
WBaseScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        try {
            int pid         = getType().getCurrent();
            GUniverse& univ = *getType().getUniverse();
            GPlanet& me     = univ.getPlanet(pid);
            switch (e.key.code) {
             case 'x':
                doBaseXchg(*this);
                return true;
             case 'f':
                doPlanetFCodeChange(univ, me);
                return true;
             case 'c':
             case 'u':
                doPlanetTransfer(univ, me, e.key.code == 'u');
                return true;
             case SDLK_F8:
                stopWithScreenId(WScreenId(WScreenId::PlanetScreen, pid));
                return true;
             case 'm':
                doBaseChangeMission(univ, me);
                return true;
             case ss_Ctrl + 'r':
                if (int sid = me.getBaseShipyardId())
                    stopWithScreenId(WScreenId(WScreenId::ShipScreen, sid));
                return true;
             case ss_Ctrl + 'c':
                if (int sid = me.getShipCloningHere(univ))
                    stopWithScreenId(WScreenId(WScreenId::ShipScreen, sid));
                return true;
             case 't':
                doTechUpgradeDialog(univ, me);
                return true;
             case 'd':
                doPlanetBuildScreen(univ, me, 2);
                return true;
             case 'a':
                doBaseBuildAmmo(univ, me);
                return true;
             case 'b':
             case SDLK_TAB:
             {
                 WShipBuildScreenWithActor bs(me);
                 ui_root->centerWidget(bs);
                 ui_root->add(bs);
                 ui_root->run();
                 ui_root->close(bs);
                 univ.doScreenUpdates();
                 return true;
             }
             case SDLK_F9:
                doEditComment(me, getPlanetProperties(me.getId()), _("Enter new comment for this planet:"), _("Edit Comment"));
                return true;
             case SDLK_INSERT:
                doPlanetAddToSim(me);
                return true;
             /* base screen keys:
                's'    ship cost
                alt-r  reset
                f10    choose fleet */

            }
        }
        catch (const GError& e) {
            messageBox(e.getUserError(), _("Error"));
            return true;
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

void
WBaseScreen::update()
{
    if (GMapObject* mo = dynamic_cast<GMapObject*>(&type.getObject(type.getCurrent())))
        getScanner().getChart()->setCenterPosition(mo->getPos());
}

WBaseScreenHeaderTile::WBaseScreenHeaderTile(GObjectType& type)
    : WControlScreenHeaderTile(type, bAuto + bAdd)
{ }

string_t
WBaseScreenHeaderTile::getTitle()
{
    return getObject()->getName();
}

string_t
WBaseScreenHeaderTile::getSubtitle()
{
    string_t result;
    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
        int16 level = p->unit_scores.getScore(planet_score_definitions.lookupScore(ScoreId_ExpLevel));
        const char* fmt = (level >= 0
                           ? _("(Id #%d, %s)")
                           : _("(Id #%d)"));
        result = format(fmt, getObject()->getId(), getExperienceLevelName(level));
    }
    return result;
}

ResId
WBaseScreenHeaderTile::getPictureId()
{
    int max_tech = 0;
    int id = 0;
    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
        id = p->getId();
        if (p->hasBase()) {
            max_tech = std::max(std::max(p->getBaseTech(tl_Engine),
                                         p->getBaseTech(tl_Hull)),
                                std::max(p->getBaseTech(tl_Beam),
                                         p->getBaseTech(tl_Torpedo)));
        }
    }
    return ResId(res::base, max_tech, id);
}


WScreenId
doBaseScreen()
{
    return WBaseScreen().runScreen();
}
