/**
  *  \file client/scr-planet.cc
  *  \brief Planet Screen
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions.h"
#include "client/cscreen.h"
#include "client/dlg-list.h"
#include "client/marks.h"
#include "client/screens.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/objlist.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/font.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/dlg-tech.h"
#include "ui/layout.h"
#include "client/dlg-structure.h"
#include "client/dlg-tax.h"
#include "ui/table.h"
#include "client/dlg-supplies.h"
#include "game/action.h"
#include "resmgr/resid.h"
#include "ui/util.h"
#include "game/tables.h"
#include "client/wscanchart.h"
#include "int/intglobal.h"
#include "int/data.h"
#include "client/dlg-comment.h"
#include "game/config.h"

namespace {
    class WPlanetSelect : public WObjectSelectionTile {
     public:
        WPlanetSelect(GPlanetType& type);
        void drawContent(GfxCanvas& can);
    };

    class WPlanetScreen : public WControlScreen {
     public:
        WPlanetScreen();
        ~WPlanetScreen();
        void update();
        bool handleEvent(const UIEvent& e, bool second_pass);
    };

    class WPlanetScreenHeaderTile : public WControlScreenHeaderTile {
     public:
        WPlanetScreenHeaderTile(GObjectType& type);
        string_t getTitle();
        string_t getSubtitle();
        ResId getPictureId();
    };

    class WPlanetEconomyTile : public WObjectTypeObserverTile {
        void drawPlanetEconomyTile(UIAbstractTableWriter& w, bool all_cols);
        ObjectHolder h;
     public:
        WPlanetEconomyTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };

    class WPlanetColonistTile : public WObjectTypeObserverTile {
        void drawPlanetColonistTile(UIAbstractTableWriter& w, bool all_cols);
        ObjectHolder h;
     public:
        WPlanetColonistTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };

    class WPlanetNativeTile : public WObjectTypeObserverTile {
     public:
        WPlanetNativeTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };
}

/***************************** WPlanetSelect *****************************/

/** \class WPlanetSelect
    \brief Planet Selector

    This widget can be used to select a played planet from a game
    turn. */

WPlanetSelect::WPlanetSelect(GPlanetType& type)
    : WObjectSelectionTile(type)
{ }

void
WPlanetSelect::drawContent(GfxCanvas& can)
{
    GPlanet& pp = getType().getUniverse()->getPlanet(getCurrent());
    UISkin& sk = getSkin();
    sk.startDraw();

    // FIXME: this is very incomplet
    int y = getExtent().y, x = getExtent().x, rem = 0;
    if (pp.isMarked()) {
        sk.drawBackground(can, GfxRect(x, y, 15, font_heights[FONT_NORMAL]));
        sk.getParameters().color = standard_colors[COLOR_GREENBLACK];
        drawSelection(can, sk, extent.x + 5, extent.y + 7, 1, 2);
        rem += 15;
    }
    sk.setColor(UIColor::tc_Static);
    sk.outTextF(can, FONT_NORMAL, x+rem, y, extent.w - rem, pp.getName());
    y += font_heights[FONT_NORMAL]+2;

    /* location, climate */
    sk.outTextF(can, FONT_NORMAL, x, y, 130, format(_("Climate: %s"), getTemperatureName(pp.getTemperature())));
    if (int sector = pp.getPos().getSectorNumber())
        sk.outTextF(can, FONT_NORMAL, x+130, y, extent.w - 130,
                    format(_("Location: (%d,%d) - Sector %d"),
                           pp.getPos().x, pp.getPos().y, sector));
    else
        sk.outTextF(can, FONT_NORMAL, x+130, y, extent.w - 130,
                    format(_("Location: (%d,%d)"),
                           pp.getPos().x, pp.getPos().y));
    y += font_heights[FONT_NORMAL]+2;

    // FIXME: I don't really like this because it uses hard
    // measurements (the same as in PCC 1.x)
    /*  |-35-|---70---|5|---95---|30|----175--------|
        Col:  N,NNN,NNN Race      xx% -> happy (100)
        Nat:  N,NNN,NNN Race
        |0   |35      |105|110   |205|235  */
    
    /* colonists */
    sk.outTextF(can, FONT_NORMAL, x,   y, 35, _("Col:"));
    if (pp.getOwner()) {
        sk.outTextF(can, FONT_NORMAL, x+110, y, 95, getAdjName(pp.getOwner()));
        sk.outTextF(can, FONT_NORMAL, x+235, y, 155,
                    format(" -> %s (%d)", getHappinessName(pp.getColonistHappiness()),
                           pp.getColonistHappiness()));
        sk.getParameters().just_x = 2;
        sk.outTextF(can, FONT_NORMAL, x+105, y, 70, clansToString(pp.getColonists()));
        sk.outTextF(can, FONT_NORMAL, x+235, y, 30, format("%d%%", pp.getColonistTax()));
        sk.getParameters().just_x = 0;
    } else {
        sk.outTextF(can, FONT_NORMAL, x+35, y, 70+5+95+30+155, _("none"));
    }
    y += font_heights[FONT_NORMAL];

    /* natives */
    sk.outTextF(can, FONT_NORMAL, x, y, 35, _("Nat:"));
    if (pp.getNativeRace()) {
        sk.outTextF(can, FONT_NORMAL, x+110, y, 95, getNativeRaceName(pp.getNativeRace()));
        sk.outTextF(can, FONT_NORMAL, x+235, y, 155,
                    format(" -> %s (%d)", getHappinessName(pp.getNativeHappiness()),
                           pp.getNativeHappiness()));
        sk.getParameters().just_x = 2;
        sk.outTextF(can, FONT_NORMAL, x+105, y, 70, clansToString(pp.getNatives()));
        sk.outTextF(can, FONT_NORMAL, x+235, y, 30, format("%d%%", pp.getNativeTax()));
        sk.getParameters().just_x = 0;
    } else {
        sk.outTextF(can, FONT_NORMAL, x+35, y, 70+5+95+30+155, _("none"));
    }
    y += font_heights[FONT_NORMAL]+2;

    /* industry */
    sk.outTextF(can, FONT_NORMAL, x, y, extent.w,
                format(_("Mines: %d; Factories: %d; Defense Posts: %d"),
                       pp.getStructures(str_Mines),
                       pp.getStructures(str_Factories),
                       pp.getStructures(str_Defense)));
    y += font_heights[FONT_NORMAL];

    sk.outTextF(can, FONT_NORMAL, x, y, extent.w,
                format(_("%d mc; %d kt supplies"),
                       pp.getCargo(el_Money), pp.getCargo(el_Supplies)));
    y += font_heights[FONT_NORMAL];

    /* minerals */
    sk.outTextF(can, FONT_NORMAL, x, y, 50, _("Mined:"));
    sk.getParameters().just_x = 2;
    sk.outTextF(can, FONT_NORMAL, x+130, y, 80, format("%dN", pp.getCargo(el_Neutronium)));
    sk.outTextF(can, FONT_NORMAL, x+210, y, 80, format("%dT", pp.getCargo(el_Tritanium)));
    sk.outTextF(can, FONT_NORMAL, x+290, y, 80, format("%dD", pp.getCargo(el_Duranium)));
    sk.outTextF(can, FONT_NORMAL, x+370, y, 80, format("%dM", pp.getCargo(el_Molybdenum)));
    y += font_heights[FONT_NORMAL];
    
    sk.outTextF(can, FONT_NORMAL, x+130, y, 80, format("%dN", pp.getOreGround(el_Neutronium)));
    sk.outTextF(can, FONT_NORMAL, x+210, y, 80, format("%dT", pp.getOreGround(el_Tritanium)));
    sk.outTextF(can, FONT_NORMAL, x+290, y, 80, format("%dD", pp.getOreGround(el_Duranium)));
    sk.outTextF(can, FONT_NORMAL, x+370, y, 80, format("%dM", pp.getOreGround(el_Molybdenum)));
    sk.getParameters().just_x = 0;
    sk.outTextF(can, FONT_NORMAL, x, y, 50, _("Ground:"));
    y += font_heights[FONT_NORMAL] + 2;

    sk.outTextF(can, FONT_NORMAL, x, y, extent.w,
                format(_("Friendly Code: %s"), pp.getFCode()));

    can.update(getExtent());
}

/** Choose a planet. Displays the dialog if necessary and lets the
    user choose a planet.
    \return planet Id, zero if no planet or cancelled */
int
choosePlanet()
{
    int nplanets = global_planet_type.getCount();
    if (nplanets == 0) {
        messageBox(_("You do not have any planets. "
                     "Unload colonists from a starship to a planet to colonize it."),
                   _("Planets"));
        return 0;
    } else if (nplanets == 1) {
        return global_planet_type.getCurrent();
    } else {
        return WPlanetSelect(global_planet_type).doStandardDialog(_("Select Planet"));
    }
}

/*************************** WPlanetEconomyTile **************************/

/** \class WPlanetEconomyTile
    \brief Planet Economy Tile

    Displays planetary economy and resources.
    <pre>
    Mineral Mines:     XX
    Factories:         XX
    Defense Posts:     XX
    Neutronium:        XX kt
    Tritanium:         XX kt
    Duranium:          XX kt
    Molybdenum:        XX kt
    Supplies:          XX kt
    </pre> */

WPlanetEconomyTile::WPlanetEconomyTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, font_heights[FONT_NORMAL] * 8), type)
{
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(0,  30, 20, 20), h.add(new UIButton("B", 'b')));
    addControl(GfxRect(25, 30, 20, 20), h.add(new UIButton("M", 'm')));
    addControl(GfxRect(0,  55, 20, 20), h.add(new UIButton("S", 's')));
    addControl(GfxRect(25, 55, 20, 20), h.add(new UIButton("D", 'd')));
    addControl(GfxRect(25, 90, 20, 20), h.add(new UIButton("C", 'c')));
}

void
WPlanetEconomyTile::drawPlanetEconomyTile(UIAbstractTableWriter& w, bool all_cols)
{
    w.newLColumn();
    w.writeFixedCell(UIColor::tc_Static, _("Mineral Mines:"));
    w.writeFixedCell(UIColor::tc_Static, _("Factories:"));
    w.writeFixedCell(UIColor::tc_Static, _("Defense Posts:"));
    w.writeFixedCell(UIColor::tc_Static, _("Neutronium:"));
    w.writeFixedCell(UIColor::tc_Static, _("Tritanium:"));
    w.writeFixedCell(UIColor::tc_Static, _("Duranium:"));
    w.writeFixedCell(UIColor::tc_Static, _("Molybdenum:"));
    w.writeFixedCell(UIColor::tc_Static, _("Supplies:"));
    if (!all_cols)
        return;

    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
        w.newRColumn(6*8);
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getStructures(str_Mines)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getStructures(str_Factories)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getStructures(str_Defense)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getCargo(el_Neutronium)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getCargo(el_Tritanium)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getCargo(el_Duranium)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getCargo(el_Molybdenum)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getCargo(el_Supplies)));
    
        w.newLColumn();
        w.writeFixedCell(UIColor::tc_Green, "");
        w.writeFixedCell(UIColor::tc_Green, "");
        w.writeFixedCell(UIColor::tc_Green, "");
        w.writeFixedCell(UIColor::tc_Green, "kt");
        w.writeFixedCell(UIColor::tc_Green, "kt");
        w.writeFixedCell(UIColor::tc_Green, "kt");
        w.writeFixedCell(UIColor::tc_Green, "kt");
        w.writeFixedCell(UIColor::tc_Green, "kt");
    }

    w.finish();
}

void
WPlanetEconomyTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL);
    drawPlanetEconomyTile(tw, true);
}

/************************** WPlanetColonistTile **************************/

/** \class WPlanetColonistTile
    \brief Planet Colonists Tile

    Displays information about the planetary colony.

    <pre>
    Population:    XX
    Tax Rate:      XX%
    Colonists are XYZ.
    </pre> */

WPlanetColonistTile::WPlanetColonistTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 3 * font_heights[FONT_NORMAL]), type)
{
    setLayout(RightJustified, 0, true);
    addControl(GfxRect(0, 3 * font_heights[FONT_NORMAL] - 20, 20, 20), h.add(new UIButton("T", 't')));
}

void
WPlanetColonistTile::drawPlanetColonistTile(UIAbstractTableWriter& w, bool /*all_cols*/)
{
    w.newLColumn();
    w.writeFixedCell(UIColor::tc_Static, _("Population:"));
    w.writeFixedCell(UIColor::tc_Static, _("Tax Rate:"));
    w.writeFixedCell(UIColor::tc_Static, _("Colonists are"));

    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
        w.newLColumn(200);
        w.writeVariableCell(UIColor::tc_Green, clansToString(p->getColonists()));
        w.writeVariableCell(UIColor::tc_Green, format("%d%%", p->getColonistTax()));

        int happy = p->getColonistHappiness();
        UISkin::Color c = (happy <= 40
                           ? UIColor::tc_Red
                           : happy < 70
                           ? UIColor::tc_Yellow
                           : UIColor::tc_Green);
        w.writeVariableCell(c, format("%s (%d)", getHappinessName(happy), happy));
    }

    w.finish();
}

void
WPlanetColonistTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 20), FONT_NORMAL);
    drawPlanetColonistTile(tw, true);
}

/*************************** WPlanetNativeTile ***************************/

/** \class WPlanetNativeTile
    \brief Planet Native Tile

    Displays information about a planet's natives, if any. */

WPlanetNativeTile::WPlanetNativeTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 4 * font_heights[FONT_NORMAL]), type)
{ }

void
WPlanetNativeTile::drawData(GfxCanvas& can)
{
    GPlanet* p = dynamic_cast<GPlanet*>(getObject());
    if (!p)
        return;

    const int y = font_heights[FONT_NORMAL];

    setColor(UIColor::tc_Static);

    if (!p->getNativeRace()) {
        GfxRect r = getInnerSize(5, 5);
        getSkin().setSolid(false);
        getSkin().getParameters().font = fonts[FONT_NORMAL];
        drawBackground(can, GfxRect(5, 0, extent.w-5, 4*y));
        outTextFormattedF(can, getSkin(), r.x, r.y, r.w, 2,
                          _("There are no natives on this planet."));
    } else {
        // This erases the sentence from above because the table
        // writer doesn't (can't?) do that
        drawBackground(can, GfxRect(5, 0, extent.w-5, 2*y));

        UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL);

        tw.newLColumn();
        tw.writeFixedCell(UIColor::tc_Static, _("Race:"));
        tw.writeFixedCell(UIColor::tc_Static, _("Population:"));
        tw.writeFixedCell(UIColor::tc_Static, _("Tax Rate:"));
        tw.writeFixedCell(UIColor::tc_Static, _("Natives are"));

        tw.newLColumn(200);
        tw.writeVariableCell(UIColor::tc_Green, string_t(getNativeRaceName(p->getNativeRace())) + ", " + getNativeGovernmentName(p->getNativeGovernment()));
        tw.writeVariableCell(UIColor::tc_Green, clansToString(p->getNatives()));
        tw.writeVariableCell(UIColor::tc_Green, format("%d%%", p->getNativeTax()));

        int happy = p->getNativeHappiness();
        UISkin::Color c = (happy <= 40
                           ? UIColor::tc_Red
                           : happy < 70
                           ? UIColor::tc_Yellow
                           : UIColor::tc_Green);
        tw.writeVariableCell(c, format("%s (%d)", getHappinessName(happy), happy));

        tw.finish();
    }
}

/**************************** WPlanetFCodeTile ***************************/

/** \class WPlanetFCodeTile
    \brief Planet FCode & Money Tile

    Displays the planet's FCode and cash. */

WPlanetFCodeTile::WPlanetFCodeTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 4*font_heights[FONT_NORMAL]), type)
{
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(0, 0, 20, 20),      h.add(new UIButton("F", 'f')));
    addControl(GfxRect(0, 4*font_heights[FONT_NORMAL]-20, 30, 20), h.add(new UIButton("F9", SDLK_F9)));
}

void
WPlanetFCodeTile::drawData(GfxCanvas& can)
{
    const int y = font_heights[FONT_NORMAL];

    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
        setColor(UIColor::tc_Static);
        outText(can, FONT_NORMAL, 5, 0, _("Funds:"));
        outText(can, FONT_NORMAL, 5, y, _("Friendly Code:"));

        setColor(UIColor::tc_Green);
        outTextF(can, FONT_NORMAL, 65, 0, 170, format("%d mc", p->getCargo(el_Money)));
        outTextF(can, FONT_NORMAL, 125, y, 70, p->getFCode());

        string_t comment;
        if (IntValue* v = getPlanetProperties(p->getId())[pp_Comment])
            comment = v->toString(false);
        if (comment.length()) {
            setColor(UIColor::tc_Yellow);
            outTextF(can, FONT_NORMAL, 5, 5*y/2, extent.w-35, comment);
        } else {
            setColor(UIColor::tc_Static);
            outTextF(can, FONT_NORMAL, 5, 5*y/2, extent.w-35, _("(press [F9] to add a note)"));
        }
    }
}

/**************************** WPlanetBaseTile ****************************/

/** \class WPlanetBaseTile
    \brief Planet/Base Tile

    On a planet, displays whether there is a base, and offers to build
    one. On a starbase, offers to go to the planet. Also offers to
    scan the planet. */

/** Constructor. \param on_base true if we are on the starbase screen,
    false if on planet */
WPlanetBaseTile::WPlanetBaseTile(GObjectType& type, bool on_base)
    : WObjectTypeObserverTile(GfxPoint(315, font_heights[FONT_TITLE]*2 + 2), type),
      on_base(on_base)
{
    setLayout(RightJustified, 0, false);
    const int he = font_heights[FONT_TITLE];
    addControl(GfxRect(0, 0, 30, he),    h.add(new UIButton("F5", SDLK_F5)));
    addControl(GfxRect(0, he+2, 30, he), h.add(new UIButton("F8", SDLK_F8)));
}

void
WPlanetBaseTile::drawData(GfxCanvas& can)
{
    GPlanet* p = dynamic_cast<GPlanet*>(getObject());
    if (!p)
        return;
    
    const int w = getExtent().w - 35;
    const char* s;

    setColor(UIColor::tc_Heading);
    getSkin().getParameters().just_x = 2;
    outTextF(can, FONT_TITLE, w, 0, w, _("Planet overview"));

    if (on_base) {
        s = _("Go to planet");
    } else if (p->hasBase()) {
        s = _("Go to starbase");
    } else if (p->isBuildingBase()) {
        setColor(UIColor::tc_Green);
        s = _("Starbase being built...");
    } else {
        s = _("Build starbase");
    }

    outTextF(can, FONT_TITLE, w, font_heights[FONT_TITLE]+2, w, s);
}

/***************************** WPlanetScreen *****************************/

/** \class WPlanetScreen
    \brief Planet Control Screen

    Standard control screen for a player's planet. */

/** Constructor. */
WPlanetScreen::WPlanetScreen()
    : WControlScreen(global_planet_type, WScreenId::PlanetScreen)
{
    addTile(holder.add(new WPlanetScreenHeaderTile(global_planet_type)));
    addTile(holder.add(new WPlanetEconomyTile(global_planet_type)),
            _("Economy:"));
    addTile(holder.add(new WPlanetNativeTile(global_planet_type)),
            _("Natives:"));
    addTile(holder.add(new WPlanetColonistTile(global_planet_type)),
            _("Colonists:"));
    addTile(holder.add(new WPlanetFCodeTile(global_planet_type)));
    addTile(holder.add(new WPlanetBaseTile(global_planet_type, false)));
    addChart(holder.add(new WScannerChartWidget()));
}

WPlanetScreen::~WPlanetScreen()
{ }

void
WPlanetScreen::update()
{
    if (GMapObject* mo = dynamic_cast<GMapObject*>(&type.getObject(type.getCurrent())))
        getScanner().getChart()->setCenterPosition(mo->getPos());
}

static void
doPlanetXchg(WPlanetScreen& scr)
{
    WScanner& scanner = scr.getScanner();
    int pid = scanner.getScannedPlanetId(scanner.getScannedShipId() != 0);
    if (pid && pid != scr.getType().getCurrent() && scr.getType().isExistant(pid))
        scr.stopWithScreenId(WScreenId(WScreenId::PlanetScreen, pid));
    else
        putKey(ss_Ctrl + SDLK_F1);
}

bool
WPlanetScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        try {
            const int pid = getType().getCurrent();
            GUniverse& univ = *getType().getUniverse();
            GPlanet& planet = univ.getPlanet(pid);
            switch(e.key.code) {
             case 'x':
                doPlanetXchg(*this);
                return true;
             case 'c':
             case 'u':
                doPlanetTransfer(univ, planet, e.key.code == 'u');
                return true;
             case 'f':
                doPlanetFCodeChange(univ, planet);
                return true;
             case 'b':
             case SDLK_TAB:
                doPlanetBuildScreen(univ, planet, 0);
                return true;
             case 'm':
                doPlanetBuildScreen(univ, planet, 1);
                return true;
             case 'd':
                doPlanetBuildScreen(univ, planet, 2);
                return true;
             case 't':
                doTaxation(univ, planet);
                return true;
             case 's':
                WSellSuppliesDialog::doDialog(univ, planet);
                return true;
             case ss_Ctrl + 's':
                WUndoSellSuppliesDialog::doDialog(getDisplayedTurn() /* FIXME */, pid);
                return true;
             case SDLK_F8:
                if (planet.hasBase()) {
                    stopWithScreenId(WScreenId(WScreenId::BaseScreen, pid));
                } else {
                    doPlanetBuildBase(univ, planet);
                }
                return true;
             case SDLK_F9:
                doEditComment(planet, getPlanetProperties(planet.getId()), _("Enter new comment for this planet:"), _("Edit Comment"));
                return true;
             case SDLK_INSERT:
                doPlanetAddToSim(planet);
                return true;
                /* planet screen keys:
                   i        mail
                   alt-r    reset
                   g        give */
            }
        }
        catch (const GError& e) {
            messageBox(e.getUserError(), _("Error")); // FIXME
            return true;
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

WPlanetScreenHeaderTile::WPlanetScreenHeaderTile(GObjectType& type)
    : WControlScreenHeaderTile(type, bAuto + bAdd + bSend)
{ }

string_t
WPlanetScreenHeaderTile::getTitle()
{
    return getObject()->getName();
}

string_t
WPlanetScreenHeaderTile::getSubtitle()
{
    string_t result;
    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
        int16 level = p->unit_scores.getScore(planet_score_definitions.lookupScore(ScoreId_ExpLevel));
        const char* fmt = (level >= 0
                           ? _("(Id #%d, %s - %dF, %s)")
                           : _("(Id #%d, %s - %dF)"));
        result = format(fmt,
                        p->getId(),
                        getTemperatureName(p->getTemperature()),
                        p->getTemperature(),
                        getExperienceLevelName(level));
    }
    return result;
}

ResId
WPlanetScreenHeaderTile::getPictureId()
{
    ResId result;
    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject()))
        result = ResId(res::planet, p->getTemperature(), p->getId());
    return result;
}

WScreenId
doPlanetScreen()
{
    return WPlanetScreen().runScreen();
}
