/**
  *  \file client/scr-ship.cc
  *  \brief Starship Control Screen
  *
  *  This contains the ship screen tile widgets, and the main driver
  *  code.
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/cscreen.h"
#include "client/dlg-cargo.h"
#include "client/dlg-fcode.h"
#include "client/dlg-list.h"
#include "client/dlg-mission.h"
#include "client/marks.h"
#include "client/screens.h"
#include "cpluslib/format.h"
#include "game/fcode.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/player.h"
#include "game/ship.h"
#include "game/spec.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "util/misc.h"
#include "util/nls.h"
#include "client/dlg-warp.h"
#include "game/mission.h"
#include "game/config.h"
#include "resmgr/resid.h"
#include "client/wscanchart.h"
#include "client/wcolorframe.h"
#include "game/univ.h"
#include "int/intglobal.h"
#include "int/data.h"
#include "client/dlg-comment.h"
#include "client/actions.h"
#include "game/host.h"

namespace {
    // FIXME?
    enum { cm_Ship_BaseMenu = cm_ControlScreenFirst + 2001 };

    /// Starship selection window.
    class WShipSelect : public WObjectSelectionTile {
     public:
        WShipSelect(GShipType& type);
        ~WShipSelect();
        void drawContent(GfxCanvas& can);
    };

    /// Ship scanner (with waypoint etc).
    class WShipScannerChartWidget : public WScannerChartWidget {
     protected:
        void  lockQueryLocation(bool& hyperjumping, int& current_gravity_pid);
        int32 lockQueryDistance(GPoint pt);
        
     public:
        WShipScannerChartWidget();
        ~WShipScannerChartWidget();
        void drawScanner(GfxCanvas& can);
        // void update(int);
        void userMoveScannerTo(GPoint pt);
    };

    /// Starship screen.
    class WShipScreen : public WControlScreen {
        WShipScannerChartWidget scanner_chart;

        void doShipBaseMenu(GShip& ship, UIBaseWidget& button);
     public:
        WShipScreen();
        ~WShipScreen();
        bool handleEvent(const UIEvent& e, bool second_pass);
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        void update();
        void setWaypointToScanner();
    };

    class WShipScreenHeaderTile : public WControlScreenHeaderTile {
     public:
        WShipScreenHeaderTile(GObjectType& type);
        string_t getTitle();
        string_t getSubtitle();
        ResId getPictureId();
    };

    /// Ship equipment tile (crew/arms/damage).
    class WShipEquipmentTile : public WObjectTypeObserverTile {
        void drawShipEquipmentTile(UIAbstractTableWriter& tw);
     public:
        WShipEquipmentTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };

    /// Ship cargo tile (ore/clans/cash).
    class WShipCargoTile : public WObjectTypeObserverTile {
        void drawCargoBar(GfxCanvas& can, int x, int y, int wi, int have, int max, int color, const char* fmt);
        void drawShipCargoTile(UIAbstractTableWriter& tw);
        ObjectHolder h;
     public:
        WShipCargoTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };

    /// Ship mission tile (mission/PE/FC).
    class WShipMissionTile : public WObjectTypeObserverTile {
        ObjectHolder h;
        WColorFrame cf;
     public:
        WShipMissionTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };

    /// Ship movement tile (waypoint/speed/mass/ETA).
    class WShipMovementTile : public WObjectTypeObserverTile {
        void drawShipMovementTile(UIAbstractTableWriter& tw, bool fixed_only);
        ObjectHolder h;
     public:
        WShipMovementTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };
}

static int acquireWaypointPerm(const string_t& reason, GShip& ship);

WTile*
createShipScreenHeaderTile(GObjectType& type)
{
    return new WShipScreenHeaderTile(type);
}

/*************************** WShipSelect **************************/

WShipSelect::WShipSelect(GShipType& type)
    : WObjectSelectionTile(type)
{ }

WShipSelect::~WShipSelect()
{ }

void
WShipSelect::drawContent(GfxCanvas& can)
{
    GUniverse& univ = *getType().getUniverse();
    GShip& ps = univ.getShip(getCurrent());
    UISkin& sk = getSkin();

    // FIXME: this is very incomplet
    int y = getExtent().y, x = getExtent().x, rem = 0;
    if (ps.isMarked()) {
        sk.drawBackground(can, GfxRect(x, y, 15, font_heights[FONT_NORMAL]));
        sk.getParameters().color = standard_colors[COLOR_GREENBLACK];
        drawSelection(can, sk, extent.x + 5, extent.y + 7, 1, 2);
        rem += 15;
    }
    sk.setColor(UIColor::tc_Static);
    sk.outTextF(can, FONT_NORMAL, x+rem, y, extent.w - rem,
                ps.getFullName() + ", " + ps.getHull().getName());
    y += font_heights[FONT_NORMAL] * 3/2;
    sk.outTextF(can, FONT_NORMAL, x, y, extent.w,
                _("Location: ") + univ.getLocationName(ps.getPos(), locs_WW | locs_Orbit | locs_Verbose));
    y += font_heights[FONT_NORMAL];
    sk.outTextF(can, FONT_NORMAL, x, y, extent.w,
                _("Waypoint: ") + univ.getLocationName(ps.getWaypoint(), locs_WW | locs_Verbose));
    y += font_heights[FONT_NORMAL];
    sk.outTextF(can, FONT_NORMAL, x, y, extent.w,
                format(_("Distance: %.2f ly"), distFromDX(ps.getWaypointDX(),
                                                          ps.getWaypointDY())));

    can.update(getExtent());
}

int
chooseShip()
{
    int nships = global_ship_type.getCount();
    if (nships == 0) {
        messageBox(_("You do not have any ships. "
                     "To build ships, use [B] on a starbase."),
                   _("Starships"));
        return 0;
    } else if (nships == 1) {
        return global_ship_type.getCurrent();
    } else {
        return WShipSelect(global_ship_type).doStandardDialog(_("Select Ship"));
    }
}

/*************************** WShipEquipmentTile **************************/

WShipEquipmentTile::WShipEquipmentTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 6*font_heights[FONT_NORMAL]+3), type)
{ }

void
WShipEquipmentTile::drawShipEquipmentTile(UIAbstractTableWriter& tw)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Engines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Primary:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Secondary:"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    tw.writeFixedCell(UIColor::tc_Static, _("Damage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Crew:"));

    tw.newLColumn();

    const GShip* s = dynamic_cast<GShip*>(getObject());
    if (!s)
        return;

    const GShip& ship = *s;
    const GHull& hull = ship.getHull();

    /* engines */
    {
        string_t en = getEngine(ship.getEngine()).getName();
        if (ship.canDoSpecial(hf_Gravitonic))
            en = format(_("Grav. %s"), en);
        if (hull.getNumEngines() > 1)
            en = format("%d x %s", hull.getNumEngines(), en);
        tw.writeVariableCell(UIColor::tc_Green, en);
    }

    /* primary */
    if (ship.getBeamCount() && ship.getBeamType()) {
        string_t bm = getBeam(ship.getBeamType()).getName();
        if (ship.getBeamCount() != 1)
            bm = format("%d x %s", ship.getBeamCount(), bm);
        tw.writeVariableCell(UIColor::tc_Green, bm);
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
    }

    /* secondary */
    if (ship.getTorpType() && ship.getTorpLauncherCount()) {
        string_t tl = format(_("%d x %s launcher%!1{s%}"),
                             ship.getTorpLauncherCount(),
                             getTorp(ship.getTorpType()).getName());
        string_t tc = (ship.getAmmo()
                       ? string_t(format(_("%d torpedo%!1{es%} aboard"), ship.getAmmo()))
                       : string_t(_("no torpedoes aboard")));
        tw.writeVariableCell(UIColor::tc_Green, tl);
        tw.writeVariableCell(UIColor::tc_Green, tc);
    } else if (ship.getBayCount()) {
        string_t tl = format(_("%d fighter bay%!1{s%}"), ship.getBayCount());
        string_t tc = (ship.getAmmo()
                       ? string_t(format(_("%d fighter%!1{s%} ready"), ship.getAmmo()))
                       : string_t(_("no fighters aboard")));
        tw.writeVariableCell(UIColor::tc_Green, tl);
        tw.writeVariableCell(UIColor::tc_Green, tc);
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
        tw.writeVariableCell(UIColor::tc_Green, string_t());
    }

    /* damage */
    if (ship.getDamage()) {
        int i = ship.getSupplies() / 5;
        int j = ship.getDamage() - i;
        UISkin::Color c = UIColor::tc_Red;
        if (j < 0) {
            j = 0;
            c = UIColor::tc_Yellow;
        }
        if (i)
            tw.writeVariableCell(c, format("%d%% (-%d: %d%%)", ship.getDamage(), i, j));
        else
            tw.writeVariableCell(c, format("%d%%", ship.getDamage()));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, _("none"));
    }

    /* crew */
    if (ship.getCrew() != hull.getCrew()) {
        tw.writeVariableCell(ship.getCrew() > hull.getCrew() ? UIColor::tc_Yellow : UIColor::tc_Red,
                             format(_("%d of %d"), ship.getCrew(), hull.getCrew()));
    } else {
        tw.writeVariableCell(UIColor::tc_Green, numToString(ship.getCrew()));
    }
    tw.finish();
}

void
WShipEquipmentTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, UITableWriter::fl_Fill);
    drawShipEquipmentTile(tw);
}

/***************************** WShipCargoTile ****************************/

WShipCargoTile::WShipCargoTile(GObjectType& type)
    // FIXME: magic numbers
    : WObjectTypeObserverTile(GfxPoint(315, 4 * font_heights[FONT_NORMAL] + 14 + 3), type)
{
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(25, 0, 20, 20), h.add(new UIButton("C", 'c')));
    addControl(GfxRect(0,  0, 20, 20), h.add(new UIButton("U", 'u')));
}

void
WShipCargoTile::drawShipCargoTile(UIAbstractTableWriter& tw)
{
    // Column 1: ore names
    // Column 2: ore amounts
    // Column 3: space
    // Column 4: rest
    // Column 5: rest amounts
    // Column 6: units

    GShip* s = dynamic_cast<GShip*>(getObject());
    if (!s)
        return;

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, "Neu:");
    tw.writeFixedCell(UIColor::tc_Static, "Tri:");
    tw.writeFixedCell(UIColor::tc_Static, "Dur:");
    tw.writeFixedCell(UIColor::tc_Static, "Mol:");

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("9999 kt"));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s->getCargo(el_Neutronium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s->getCargo(el_Tritanium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s->getCargo(el_Duranium)));
    tw.writeVariableCell(UIColor::tc_Green, format("%d kt", s->getCargo(el_Molybdenum)));

    tw.newLColumn(5);           // spacer

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Colonists:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Supplies:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Money:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Total Mass:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("999,999"));
    tw.writeVariableCell(UIColor::tc_Green, clansToString(s->getCargo(el_Colonists)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(s->getCargo(el_Supplies)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(s->getCargo(el_Money)));
    tw.writeVariableCell(UIColor::tc_Green, itoa(s->getMass()));

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Green, "");
    tw.writeFixedCell(UIColor::tc_Green, "kt");
    tw.writeFixedCell(UIColor::tc_Green, "mc");
    tw.writeFixedCell(UIColor::tc_Green, "kt");

    tw.finish();
}

void
WShipCargoTile::drawData(GfxCanvas& can)
{
    GShip* s = dynamic_cast<GShip*>(getObject());
    if (!s)
        return;

    GfxRect pos = getInnerSize(0, 0);
    const int y = font_heights[FONT_NORMAL];

    UITableWriter tw(can, getSkin(), getInnerSize(5, 0), FONT_NORMAL, 0);
    drawShipCargoTile(tw);

    int cargo = s->getHull().getCargo();
    drawCargoBar(can, pos.x + 5,   pos.y + 4*y + 4, 130, s->getCargo(el_Neutronium), s->getHull().getFuel(), COLOR_GREENBLACK, _("Fuel: %d free"));
    drawCargoBar(can, pos.x + 145, pos.y + 4*y + 4, 130, cargo - s->getFreeCargo(), cargo, COLOR_BLUE, _("Cargo: %d free"));
}

void
WShipCargoTile::drawCargoBar(GfxCanvas& can, int x, int y, int wi, int have, int max,
                             int color, const char* fmt)
{
    drawFrameDown(can, GfxRect(x, y, wi, 10));
    ++x;
    ++y;
    wi -= 2;

    int split = max ? wi * have / max : 0;
    if (have > 0 && split == 0)
        split = 1;
    if (have >= max)
        split = wi;
    else if (split >= wi)
        --split;

    if (split)
        drawSolidBar(can, GfxRect(x, y, split, 8), standard_colors[color]);
    if (split != wi)
        drawSolidBar(can, GfxRect(x + split, y, wi-split, 8), standard_colors[COLOR_BLACK]);

    getSkin().getParameters().just_x = 1;
    getSkin().getParameters().just_y = 0;
    if (have > max)
        getSkin().getParameters().color = standard_colors[COLOR_RED];
    else
        getSkin().getParameters().color = standard_colors[COLOR_WHITE];
    getSkin().getParameters().bgcolor = GFX_TRANSPARENT;
    getSkin().getParameters().font = fonts[FONT_SMALL];
    ::outTextF(can, getSkin(), x + wi/2, y-1, wi, format(fmt, max-have));
}

/**************************** WShipMissionTile ***************************/

WShipMissionTile::WShipMissionTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 9*font_heights[FONT_NORMAL]/2), type)
{
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(27, 0, 20, 20),  h.add(new UIButton("M", 'm')));
    addControl(GfxRect(27, 25, 20, 20), h.add(new UIButton("E", 'e')));
    addControl(GfxRect(27, 50, 20, 20), h.add(new UIButton("F", 'f')));

    cf.add(h.add(new UIButton(cm_Ship_BaseMenu, "B", 'b')).addCommand());
    addControl(GfxRect(0, 48, 24, 24), cf);

    addControl(GfxRect(17, 75, 30, 20), h.add(new UIButton("F9", SDLK_F9)));
}

void
WShipMissionTile::drawData(GfxCanvas& can)
{
    const int y = font_heights[FONT_NORMAL];
    const int x = 120;

    setColor(UIColor::tc_Static);
    outText(can, FONT_NORMAL, 5,   0, _("Mission:"));
    outText(can, FONT_NORMAL, 5,   y, _("Primary Enemy:"));
    outText(can, FONT_NORMAL, 5, 2*y, _("Friendly Code:"));

    // cannot be const.
    GShip* s = dynamic_cast<GShip*>(getObject());
    if (!s)
        return;
    GShip& ship = *s;

    const GMission* msn =
        GMissionList::getInstance().getMissionByNumber(ship.getMission(),
                                                       GPlayerSet(raceId(ship.getRealOwner())));

    setColor(UIColor::tc_Green);
    if (msn)
        outTextF(can, FONT_NORMAL, x-50, 0, 214,
                 msn->getName(ship));
    else
        outTextF(can, FONT_NORMAL, x-50, 0, 214,
                 format("M%d I%d T%d", ship.getMission(), ship.getIntId(), ship.getTowId()));

    outTextF(can, FONT_NORMAL, x,    y, 164-25,
             (ship.getPE()
              ? getShortName(ship.getPE())
              : string_t(_("none"))));
    outTextF(can, FONT_NORMAL, x,  2*y, 100, ship.getFCode());

    // Set color of the 'b' button
    WColorFrame::Color theColor = WColorFrame::cNone;
    if (GUniverse* univ = getType().getUniverse()) {
        int baseId = ship.getOrbit();
        if (baseId != 0 && univ->isValidPlanetId(baseId)) {
            GPlanet& planet = univ->getPlanet(baseId);
            if (planet.isPlayed() && planet.hasBase()) {
                // Planet has a base
                if (planet.getBaseShipyardId() == ship.getId()) {
                    if (planet.getBaseShipyardOrder() == sya_Recycle)
                        theColor = WColorFrame::cRed;
                    else if (planet.getBaseShipyardOrder() == sya_Fix)
                        theColor = WColorFrame::cGreen;
                }
                // FIXME: handle cloning. PCC 1.x uses green if cloning, yellow if conflict
                // (multiple cloners)
            }
        }
    }
    cf.setColor(theColor);

    string_t comment;
    if (IntValue* v = getShipProperties(ship.getId())[sp_Comment])
        comment = v->toString(false);
    if (comment.empty()) {
        setColor(UIColor::tc_Static);
        comment = _("(press [F9] to add a note)");
    } else {
        setColor(UIColor::tc_Yellow);
    }
    outTextF(can, FONT_NORMAL, 5,  7*y/2, 265, comment);
}

/*************************** WShipMovementTile ***************************/

WShipMovementTile::WShipMovementTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 7*font_heights[FONT_NORMAL]), type)
{
    const int he = 7*font_heights[FONT_NORMAL];
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(0, he - 20, 20, 20), h.add(new UIButton("A", 'a')));
    addControl(GfxRect(0, he - 45, 20, 20), h.add(new UIButton("W", 'w')));
}

void
WShipMovementTile::drawShipMovementTile(UIAbstractTableWriter& tw, bool fixed_only)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Location:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Waypoint:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Distance:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Warp Factor:"));
    tw.writeFixedCell(UIColor::tc_Static, _("E.T.A.:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Fuel Usage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Engine Load:"));

    if (fixed_only)
        return;                 // save time by not computing movement info

    GShip* ps = dynamic_cast<GShip*>(getObject());
    if (!ps)
        return;

    /* compute stuff */
    GShip& s = *ps;
    GShipTurnPredictor crystal_ball(getDisplayedTurn().getCurrentUniverse(), s.getId());
    crystal_ball.computeMovement();

    const bool is_hyper = s.isHyperdriving();

    tw.newLColumn();

    /* location */
    tw.writeVariableCell(UIColor::tc_Green, getDisplayedTurn().getCurrentUniverse().getLocationName(s.getPos(), locs_WW | locs_Verbose));

    /* waypoint */
    if (s.getMission() == GMission::msn_Intercept
        && s.getIntId() > 0
        && s.getIntId() <= NUM_SHIPS
        && getDisplayedTurn().getShip(s.getIntId()).isExistant())
    {
        tw.writeVariableCell(UIColor::tc_Green,
                             getDisplayedTurn().getShip(s.getIntId()).getName());
    } else {
        tw.writeVariableCell(UIColor::tc_Green,
                             getDisplayedTurn().getCurrentUniverse().getLocationName(s.getWaypoint(),
                                                                locs_WW | locs_Verbose | locs_Ships));
    }

    /* distance */
    double dist = distFromDX(s.getWaypointDX(), s.getWaypointDY());
    tw.writeVariableCell(!is_hyper
                         ? UIColor::tc_Green
                         : dist < 20
                         ? UIColor::tc_Red
                         : dist < 340 || dist > 360  // FIXME
                         ? UIColor::tc_Yellow
                         : UIColor::tc_Green,
                         format(_("%.2f ly"), dist));

    /* warp */
    if (s.getWarp() == 0 && (s.getWaypointDX() | s.getWaypointDY())) {
        tw.writeVariableCell(UIColor::tc_Red, _("not moving"));
    } else {
        // Borland sucks.
        string_t st;
        if (is_hyper)
            st = _("Hyperdrive");
        else if (s.getWarp() == 0)
            st = _("not moving");
        else
            st = format(_("Warp %d"), s.getWarp());
        tw.writeVariableCell(UIColor::tc_Green, st);
    }

    /* e.t.a. */
    {
        string_t eta;
        UISkin::Color c = UIColor::tc_Green;

        if (s.getWaypointDX() == 0 && s.getWaypointDY() == 0) {
            eta = _("at waypoint");
        } else if (is_hyper) {
            eta = _("unknown"); // FIXME?
        } else if (s.getWarp() == 0) {
            c = UIColor::tc_Red;
            eta = _("not moving");
        } else if (crystal_ball.getTurnCount() >= crystal_ball.MOVEMENT_TIME_LIMIT) {
            eta = _("too long");
        } else {
            eta = format(_("%d turn%!1{s%}"), crystal_ball.getTurnCount());
        }
        tw.writeVariableCell(c, eta);
    }

    /* fuel usage */
    {
        string_t fuel;
        UISkin::Color c = UIColor::tc_Green;

        fuel = format(_("%d kt"), crystal_ball.getFuelUsed());
        if (crystal_ball.getTurnCount() >= crystal_ball.MOVEMENT_TIME_LIMIT)
            c = UIColor::tc_Yellow;
        else if (crystal_ball.getFuelUsed() > s.getCargo(el_Neutronium))
            c = UIColor::tc_Red;

        // FIXME: cloak fuel
        tw.writeVariableCell(c, fuel);
    }

    /* load */
    int towee_mass = 0;
    if (s.getMission() == GMission::msn_Tow
        && s.getTowId() > 0 && s.getTowId() <= NUM_SHIPS
        && getDisplayedTurn().getShip(s.getTowId()).isExistant())
    {
        towee_mass = getDisplayedTurn().getShip(s.getTowId()).getMass();
    }
    tw.writeVariableCell(UIColor::tc_Green,
                         format(_("%d kt"), s.getMass() + towee_mass));

    tw.finish();
}

void
WShipMovementTile::drawData(GfxCanvas& can)
{
    UITableWriter tw(can, getSkin(), getInnerSize(5, 20), FONT_NORMAL, UITableWriter::fl_Fill);
    drawShipMovementTile(tw, false);
}

/************************ WShipScannerChartWidget ************************/

WShipScannerChartWidget::WShipScannerChartWidget()
{ }

WShipScannerChartWidget::~WShipScannerChartWidget()
{ }

void
WShipScannerChartWidget::lockQueryLocation(bool& hyperjumping, int& current_gravity_pid)
{
    const GShip& sh = global_ship_type.getShip();
    hyperjumping        = sh.isHyperdriving();
    current_gravity_pid = getDisplayedTurn().getCurrentUniverse().getGravityPlanetAt(sh.getPos());
}

int32
WShipScannerChartWidget::lockQueryDistance(GPoint pt)
{
    const GShip& sh = global_ship_type.getShip();
    int32 dist2 = distanceSquared(sh.getPos(), pt);
    long foo = host.isPHost() ? 0 : 1;
    if (sh.isHyperdriving() && !(dist2-foo >= 340L*340 && dist2+foo <= 360L*360)) {
        return -1;
    } else {
        return dist2;
    }
}

void
WShipScannerChartWidget::drawScanner(GfxCanvas& can)
{
    const GChartViewport& vp = getViewport();

    GPoint                wp = global_ship_type.getShip().getWaypoint();
    GfxPoint       wp_screen = vp.scaleXY(wp);

    if (wp != getScannerPosition() && wp_screen != vp.getCenter()) {
        GfxParameters parm;
        parm.color = standard_colors[COLOR_RED];

        drawLine(can, parm, vp.getCenter().x, vp.getCenter().y, wp_screen.x, wp_screen.y);
        drawEndBig(can, parm, wp_screen);
    }

    WScannerChartWidget::drawScanner(can);
}

// void
// WShipScannerChartWidget::update(int)
// {
//     drawWidget(false);
// }

// void
// WShipScannerChartWidget::handleScannerChange(bool dblclick)
// {
//     if (dblclick) {
//         GGameTurn& trn = getDisplayedTurn();
//         if (int id = acquireWaypointPerm(_("waypoint"), trn.getShipType().getShip())) {
//             trn.getShip(id).setWaypoint(getScannerPosition());
//             trn.getCurrentUniverse().doUpdates();
//         }
//     }
// }

void
WShipScannerChartWidget::userMoveScannerTo(GPoint pt)
{
    bool scanner_tied_to_waypoint = false;
    if (scanner_tied_to_waypoint) {
        // FIXME
        setScannerPosition(pt);
    } else {
        setScannerPosition(pt);
    }
}

/****************************** WShipScreen ******************************/

WShipScreen::WShipScreen()
    : WControlScreen(global_ship_type, WScreenId::ShipScreen),
      scanner_chart()
{
    addTile(holder.add(new WShipScreenHeaderTile(global_ship_type)));
    addTile(holder.add(new WShipEquipmentTile(global_ship_type)),    _("Equipment & Crew:"));
    addTile(holder.add(new WShipCargoTile(global_ship_type)),        _("Aboard:"));
    addTile(holder.add(new WShipMissionTile(global_ship_type)),      _("Mission:"));
    addTile(holder.add(new WShipMovementTile(global_ship_type)),     _("Movement:"));
    addChart(scanner_chart);

    scanner_chart.sig_dblclick.add(this, &WShipScreen::setWaypointToScanner);
}

WShipScreen::~WShipScreen()
{ }

void
WShipScreen::update()
{
    int sid = getType().getCurrent();
    const GShip& ship = getType().getUniverse()->getShip(sid);
    scanner_chart.setCenterPosition(ship.getPos());
    scanner_chart.setScannerPosition(scanner_chart.clipScannerPosition(ship.getWaypoint()));
    scanner_chart.drawWidget();
}

static void
doShipXchg(WControlScreen& scr)
{
    WScanner& scan             = scr.getScanner();
    WScannerChartWidget* chart = scan.getChart();
    int n;
    if (scan.getScannedShipId()) {
        n = scr.getType().getUniverse()->getNumberOfPlayedShipsAt(scr.getScannerPosition());
        if (chart->getCenterPosition() == chart->getScannerPosition())
            --n;                // discount current ship
    } else {
        n = 0;
    }

    if (n)
        putKey(ss_Ctrl + SDLK_F1);
    else
        putKey(ss_Ctrl + SDLK_F2);
}

/** Attempt to acquire permission to change waypoint. (this is used for fleets)
    \param reason reason, for user
    \param ship   ship we want to change
    \returns 0 if permission not granted, ship id to change waypoint on
    specified ship */
static int
acquireWaypointPerm(const string_t& reason, GShip& ship)
{
    if (!ship.getWaypointPermission()) {
        // FIXME: display dialog, leave fleet, ...
        (void) reason;
        return 0;
    } else {
        return ship.getId();
    }
}

void
doShipTransferFor(GShip& ship, GObjectId chosen)
{
    GMultiTransaction mt;
    GShipTransfer&   left = mt.addNewSubtransaction(new GShipTransfer(ship));
    GCargoContainer* right;

    if (!chosen.isGameObject()) {
        /* jettison */
        right = &mt.addNewSubtransaction(new GShipTransporter(ship));
    } else if (GShip* s = dynamic_cast<GShip*>(&chosen.getObject())) {
        if (s->getOwner() == ship.getOwner()) {
            /* ship <-> ship */
            ASSERT(s->hasShipData());
            right = &mt.addNewSubtransaction(new GShipTransfer(*s));
        } else {
            /* ship <-> foreign ship */
            if (ship.isTransporterInUse(GShip::t_Transfer)) {
                messageBox(_("The cargo transporter is already in use."), _("Cargo Transfer"));
                return;
            }
            if (ship.getTransporter(GShip::t_Transfer).id != 0
                && ship.getTransporter(GShip::t_Transfer).id != s->getId()
                && ship.isTransporterActive(GShip::t_Transfer))
            {
                messageBox(format(_("The cargo transporter is already transferring cargo "
                                    "to ship #%d."),
                                  ship.getTransporter(GShip::t_Transfer).id), _("Cargo Transfer"));
                // FIXME: state name of current target ship here
                // FIXME: we could cancel the transfer here
                return;
            }
            // ship.cancelTranporter(true);
            right = &mt.addNewSubtransaction(new GShipTransporter(ship, GShip::t_Transfer, *s));
        }
    } else if (GPlanet* p = dynamic_cast<GPlanet*>(&chosen.getObject())) {
        if (p->getOwner() == ship.getOwner() && p->hasPlanetData()) {
            right = &mt.addNewSubtransaction(new GPlanetTransfer(*p));
        } else {
            if (ship.isTransporterInUse(GShip::t_Unload)) {
                messageBox(_("The cargo transporter is already in use."), _("Cargo Transfer"));
                return;
            }
            right = &mt.addNewSubtransaction(new GShipTransporter(ship, GShip::t_Unload, *p));
        }
    } else {
        FAIL();
        right = 0;
    }

    doCargoTransfer(mt, mt.addNewSubtransaction(new GCargoTransfer(left, *right)));
}

static void
doShipCargoTransfer(GUniverse& univ, GShip& ship)
{
    GObjectList list;
    int pid = univ.getPlanetAt(ship.getPos());
    if (pid)
        list.addObject(global_anyplanet_type, pid);
    else
        list.addNonObject(-1, _("Jettison into space"));

    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (i != ship.getId() && univ.getShip(i).isExistant() && univ.getShip(i).getPos() == ship.getPos())
            list.addObject(global_anyship_type, i);

    GObjectId chosen = WObjectSelectionList(0, 15, 320, list).doStandardDialog(_("Transfer cargo to..."), 0 /* FIXME */);
    if (chosen.isValid())
        doShipTransferFor(ship, chosen);
}

static void
doShipNavigationChart(GUniverse& /*univ FIXME*/, GShip& ship)
{
    int size = ui_root->getExtent().h;

    UIWindow the_window(_("Navigation Chart"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    the_window.setLayout(UIVBoxLayout::instance5);

    WScannerChartWidget chart;
    chart.setCenterPosition(ship.getPos());
    chart.setScannerPosition(chart.clipScannerPosition(ship.getWaypoint()));
    chart.setState(chart.sc_StyleChange | chart.sc_Zoomable | chart.sc_Keyboard, true);
    the_window.add(chart);

    UIGroup buttons(UIHBoxLayout::instance5);
    ObjectHolder h;
    buttons.add(h.add(new UISpacer()));
    buttons.add(h.add(new UIButton(cm_Enter, _("F10 - OK"), SDLK_F10)).addStop());
    buttons.add(h.add(new UIButton(cm_Escape, "ESC",        SDLK_ESCAPE)).addStop());
    the_window.add(buttons);
    the_window.setExtent(GfxRect(0, 0, size - 50, size - 10));

    ui_root->centerWidget(the_window);
    ui_root->add(the_window);
    int rv = ui_root->run();
    ui_root->close(the_window);

    if (rv == cm_Enter)
        ship.setWaypoint(chart.getScannerPosition());
}

void
WShipScreen::setWaypointToScanner()
{
    GGameTurn& trn = getDisplayedTurn();
    if (int id = acquireWaypointPerm(_("waypoint"), global_ship_type.getShip())) {
        trn.getShip(id).setWaypoint(getScannerPosition());
        setOptimumWarp(trn.getCurrentUniverse(), id);
        trn.getCurrentUniverse().doScreenUpdates();
    }
}

static void
doChangePE(GUniverse& univ, GShip& ship)
{
    UIStringList d;
    d.add(0, _("0 - none"));
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        d.add(i, format("%X - %s", i, getShortName(i)));

    UIStandardListbox lb(0, -1, 240, d);
    lb.scrollToKey(ship.getPE());
    if (lb.doStandardListbox(_("Change Primary Enemy"), 0))
        ship.setPE(lb.getKey());
    univ.doScreenUpdates();
}

void
WShipScreen::doShipBaseMenu(GShip& ship, UIBaseWidget& button)
{
    enum {
        CancelShipyard,
        FixThisShip,
        RecycleThisShip,
        GoToBase
    };

    // Are we orbiting a base?
    if (!global_base_type.isExistant(ship.getOrbit())) {
        messageBox(_("We are not at one of our starbases."), _("Starbase Commands"));
        return;
    }
    GPlanet& base = global_base_type.getObject(ship.getOrbit());
    GUniverse& univ = *global_base_type.getUniverse();

    // Build command menu
    UIStringList cmdlist;

    // - Fix/Recycle
    // FIXME: the "can we fix" logic is also in act-fix.cc; can we share it?
    int currentAction = base.getBaseShipyardId() == ship.getId() ? base.getBaseShipyardOrder() : 0;
    if (currentAction == sya_Fix) {
        cmdlist.add(CancelShipyard, _("Cancel \"repair\" order"));
    } else {
        cmdlist.add(FixThisShip, _("Fix (repair) this ship"));
    }
    if (currentAction == sya_Recycle) {
        cmdlist.add(CancelShipyard, _("Cancel \"recycle\" order"));
    } else {
        if (base.getOwner() == ship.getOwner())
            cmdlist.add(RecycleThisShip, _("Recycle this ship"));
    }

    // - Build torps/fighters FIXME

    // - Starbase screen
    cmdlist.add(GoToBase, _("Starbase screen"));

    // - Clone FIXME

    // Show menu
    UIStandardListbox lb(0, 1, cmdlist.getWidth(), cmdlist);
    if (!lb.doStandardMenu(button))
        return;

    switch (int cmd = lb.getKey()) {
     case CancelShipyard:
        base.setBaseShipyardOrder(0, 0);
        univ.doScreenUpdates();
        break;

     case FixThisShip:
     case RecycleThisShip:
        {
            bool accept = true;
            if (base.getBaseShipyardOrder() != sya_None && base.getBaseShipyardId() != ship.getId()) {
                // Find name of ship being repaired/recycled
                string_t shipName;
                if (univ.isValidShipId(base.getBaseShipyardId()) && univ.getShip(base.getBaseShipyardId()).isExistant())
                    shipName = univ.getShip(base.getBaseShipyardId()).getName();
                else
                    shipName = format(_("Ship %d"), base.getBaseShipyardId());
                // Ask user
                if (base.getBaseShipyardOrder() == sya_Fix) {
                    accept = UIMessageBox(format(_("This starbase is already repairing %s. "
                                                   "If you continue, this order will be replaced. Continue?"),
                                                 shipName),
                                          _("Starbase Commands")).doYesNoDialog();
                } else if (base.getBaseShipyardOrder() == sya_Recycle) {
                    accept = UIMessageBox(format(_("This starbase is already recycling %s. "
                                                   "If you continue, this order will be replaced. Continue?"),
                                                 shipName),
                                          _("Starbase Commands")).doYesNoDialog();
                }
            }
            if (accept) {
                base.setBaseShipyardOrder(cmd == FixThisShip ? sya_Fix : sya_Recycle,
                                          ship.getId());
                univ.doScreenUpdates();
            }
        }
        break;

     case GoToBase:
        stopWithScreenId(WScreenId(WScreenId::BaseScreen, ship.getOrbit()));
        break;
    }
}

bool
WShipScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        GUniverse& univ = *getType().getUniverse();
        GShip&     ship = global_ship_type.getShip();
        switch (e.key.code) {
         case 'n': {
             string_t s = ship.getName();
             if (UIInputLine::run(s, 20, format(_("Rename Starship #%d"),
                                                ship.getId()),
                             _("Enter new name:"), 20, 0))
                 ship.setName(s);
             univ.doScreenUpdates();
             return true;
         }
         case 'm':
            doShipMissionChange(univ, ship);
            return true;
         case 'e':
            doChangePE(univ, ship);
            return true;
         case 'a':
            if (int id = acquireWaypointPerm(_("waypoint"), ship)) {
                doShipNavigationChart(univ, univ.getShip(id));
                univ.doScreenUpdates();
            }
            return true;
         case 'c':
            doShipCargoTransfer(univ, ship);
            univ.doScreenUpdates();
            return true;
         case ss_Ctrl + 'j':
            if (univ.getPlanetAt(ship.getPos()) == 0) {
                doShipTransferFor(ship, GObjectId());
                univ.doScreenUpdates();
            }
            return true;
         case ss_Ctrl + 's':
            if (ship.getTransporter(GShip::t_Transfer).id) {
                doShipTransferFor(ship, GObjectId(global_ship_type,
                                                  ship.getTransporter(GShip::t_Transfer).id));
                univ.doScreenUpdates();
            }
            return true;
         case ss_Ctrl + 'p':
            // FIXME: should possibly also handle beam-up orders
            if (ship.getTransporter(GShip::t_Unload).id) {
                doShipTransferFor(ship, GObjectId(global_anyplanet_type,
                                                  ship.getOrbit()));
                univ.doScreenUpdates();
            }
            return true;
         case 'u':
            if (ship.getOrbit()) {
                doShipTransferFor(ship, GObjectId(global_anyplanet_type,
                                                  ship.getOrbit()));
                univ.doScreenUpdates();
            } else {
                messageBox(_("We are not orbiting a planet, sir. "
                             "If you want to jettison cargo, use the "
                             "'c' command instead."),
                           _("Cargo Transfer"));
            }
            return true;
         case 'x':
            doShipXchg(*this);
            return true;
         case 'f': {
             GFCodeList list(GFCodeList::getInstance(), ship);
             WFCodeWindow dialog(_("Change Friendly Code"), list);
             dialog.setFCode(ship.getFCode());
             ui_root->centerWidget(dialog);
             ui_root->add(dialog);
             int rv = ui_root->run();
             ui_root->close(dialog);
             if (rv == cm_Enter) {
                 ship.setFCode(dialog.getFCode());
                 univ.doScreenUpdates();
             }
             return true;
         }
         case 'w':
            if (int id = acquireWaypointPerm(_("speed"), ship)) {
                (WShipSpeedSelector(univ.getShip(id)).doStandardDialog());
            }
            return true;
         case ss_Ctrl + 'w':
            setWaypointToScanner();
            return true;
         case ss_Ctrl + 'i':
            // FIXME: if mission says arg is a planet, go to planet instead
            if (ship.getIntId())
                stopWithScreenId(WScreenId(WScreenId::ShipScreen, ship.getIntId()));
            return true;
         case ss_Ctrl + 't':
            // FIXME: if mission says arg is a planet, go to planet instead
            if (ship.getTowId())
                stopWithScreenId(WScreenId(WScreenId::ShipScreen, ship.getTowId()));
            return true;
         case SDLK_F8:
         case ss_Shift + SDLK_F8:
            stopWithScreenId(WScreenId(WScreenId::BaseScreen, ship.getOrbit()));
            return true;
         case SDLK_TAB:
         case ss_Shift + SDLK_TAB:
         case ss_Ctrl + SDLK_TAB:
         case ss_Shift + ss_Ctrl + SDLK_TAB: {
            GObjectList l;
            l.addObjectsAt(univ, ship.getPos(), 0, 0);
            l.postprocessUser(false);
            int cur = l.getIndexFor(ship);
            int n = ((e.key.code & ss_Shift)
                     ? l.getPrev(cur, (e.key.code & ss_Ctrl) != 0)
                     : l.getNext(cur, (e.key.code & ss_Ctrl) != 0));
            if (n > 0)
                stopWithScreenId(WScreenId(getScreenId(), l.getObject(n).getId()));
            return true;
         }
         case SDLK_F9:
            doEditComment(ship, getShipProperties(ship.getId()), _("Enter new comment for this ship:"), _("Edit Comment"));
            return true;
         case SDLK_INSERT:
            doShipAddToSim(ship, true);
            return true;
            /* ship screen keys:

               m-i, m-t         // search xceptor/tower
               c-n              // list next turn
               q                // toggle independantmove
               s                // spec
               r                // remote
               g                // give
               d                // distribute cargo
               f10              // fleet
               s-f10            // join fleet/goto fleet
               m-r              // reset x/y
            */
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

bool
WShipScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_Ship_BaseMenu) {
        doShipBaseMenu(global_ship_type.getShip(), sender);
        return true;
    } else {
        return WControlScreen::handleCommand(cmd, sender);
    }
}

WShipScreenHeaderTile::WShipScreenHeaderTile(GObjectType& type)
    : WControlScreenHeaderTile(type, bName + bAuto + bCycle + bAdd)
{ }

string_t
WShipScreenHeaderTile::getTitle()
{
    return getObject()->getName();
}

string_t
WShipScreenHeaderTile::getSubtitle()
{
    if (GShip* sh = dynamic_cast<GShip*>(getObject())) {
        /* Format is "(Id #%d, [race] [level] hull)".
           Since format() cannot easily suppress strings conditionally, we
           implement the suppression here. The suppressed parameters still
           appear in the format string, but the %! makes them not show up.
           This has the added benefit of making the format strings for the
           "(ID race hull)" and (ID level hull)" cases different which some
           languages may need. An alternative would have been to use the
           %nn$ syntax to move the argument pointer. */
        int16 level = sh->unit_scores.getScore(ship_score_definitions.lookupScore(ScoreId_ExpLevel));
        const char* fmt = (sh->getRealOwner() != getPlayerId()
                           ? (level >= 0
                              ? _("(Id #%d, %s %s %s)")
                              : _("(Id #%d, %s %!s%s)"))
                           : (level >= 0
                              ? _("(Id #%d, %!s%s %s)")
                              : _("(Id #%d, %!s%!s%s)")));
        return format(fmt)
            << sh->getId()
            << getAdjName(sh->getRealOwner())
            << getExperienceLevelName(level)
            << sh->getHull().getName();
    } else {
        return string_t();
    }
}

ResId
WShipScreenHeaderTile::getPictureId()
{
    if (GShip* sh = dynamic_cast<GShip*>(getObject())) {
        return ResId(res::ship, sh->getHull().getPicture(), sh->getHullId());
    } else {
        return ResId();
    }
}

WScreenId
doShipScreen()
{
    return WShipScreen().runScreen();
}
