/**
  *  \file client/scr-ufo.cc
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "game/global.h"
#include "util/nls.h"
#include "ui/layout.h"
#include "gfx/palette.h"
#include "game/formula.h"
#include "cpluslib/format.h"

namespace {
    class WUfoInfoTile : public WObjectTypeObserverTile {
     public:
        WUfoInfoTile(GObjectType& type);
        void drawData(GfxCanvas& can);
    };

    class WUfoInfoWindow : public UIWindow {
        ObjectHolder h;
        GObjectType& type;
     public:
        WUfoInfoWindow(GObjectType& type);
        bool handleEvent(const UIEvent& e, bool second_pass);
    };
}

/****************************** WUfoInfoTile *****************************/

WUfoInfoTile::WUfoInfoTile(GObjectType& type)
    : WObjectTypeObserverTile(font_sizes[FONT_NORMAL].scaledBy(25, 9),
                              type)
{ }

void
WUfoInfoTile::drawData(GfxCanvas& can)
{
    GUfo* u = dynamic_cast<GUfo*>(getObject());
    if (!u)
        return;

    // Info1
    // Info2
    // Location:     (X,Y)
    // Radius:       N ly
    // Speed:        warp x
    // Heading:      unknown
    // Visible at:   n ly (from planet)
    //               n ly (from ship)
    // Last seen:    this turn

    GfxRect pos = getInnerSize(0, 0);
    const int dy = font_heights[FONT_NORMAL];
    const int x0 = font_ems[FONT_NORMAL];
    const int x1 = font_ems[FONT_NORMAL]*13;
    const int w  = pos.w - x1;
    int y = 0;
    string_t tmp;

    setColor(UIColor::tc_Static);

    // Lines 1, 2:
    outTextF(can, FONT_NORMAL, x0, y, pos.w-x0, u->getInfo1());
    y += dy;
    outTextF(can, FONT_NORMAL, x0, y, pos.w-x0, u->getInfo2());
    y += dy;

    // Line 3:
    outText (can, FONT_NORMAL, x0, y, _("Location:"));
    outTextF(can, FONT_NORMAL, x1, y, w, format("(%d,%d)", u->getPos().x, u->getPos().y));
    y += dy;

    // Line 4:
    outText (can, FONT_NORMAL, x0, y, _("Radius:"));
    if (u->getRadius() <= 0)
        tmp = _("(small)");
    else
        tmp = format(_("%d ly"), u->getRadius());
    outTextF(can, FONT_NORMAL, x1, y, w, tmp);
    y += dy;

    // Line 5:
    outText (can, FONT_NORMAL, x0, y, _("Speed:"));
    if (u->getSpeed() <= 0)
        tmp = _("not moving");
    else
        tmp = format(_("warp %d"), u->getSpeed());
    outTextF(can, FONT_NORMAL, x1, y, w, tmp);
    y += dy;

    // Line 6:
    outText (can, FONT_NORMAL, x0, y, _("Heading:"));
    if (u->getHeading() < 0)
        tmp = _("unknown");
    else
        tmp = format("%d", u->getHeading());
    if (u->getMovementVector().x != 0 || u->getMovementVector().y != 0)
        tmp += format(" (%+d,%+d)", u->getMovementVector().x, u->getMovementVector().y);
    outTextF(can, FONT_NORMAL, x1, y, w, tmp);
    y += dy;

    // Line 7, 8:
    outText (can, FONT_NORMAL, x0, y, _("Visible at:"));
    outTextF(can, FONT_NORMAL, x1, y, w, format(_("%d ly (from planet)"), u->getPlanetRange()));
    y += dy;
    outTextF(can, FONT_NORMAL, x1, y, w, format(_("%d ly (from ship)"), u->getShipRange()));
    y += dy;

    // Line 9:
    // FIXME: make a function for this, it's duplicated everywhere
    outText (can, FONT_NORMAL, x0, y, _("Last info:"));
    int age = getDisplayedTurn().getTurnNumber() - u->getLastTurn();
    if (age == 0)
        outTextF(can, FONT_NORMAL, x1, y, w, _("this turn"));
    else if (age == 1)
        outTextF(can, FONT_NORMAL, x1, y, w, _("last turn"));
    else if (age > 0)
        outTextF(can, FONT_NORMAL, x1, y, w, format(_("%d turns ago"), age));
    else
        outTextF(can, FONT_NORMAL, x1, y, w, format(_("Turn %d"), u->getLastTurn()));
    y += dy;
}

/***************************** WUfoInfoWindow ****************************/

WUfoInfoWindow::WUfoInfoWindow(GObjectType& type)
    : UIWindow(_("Ufo Information"), BLUE_WINDOW, COLOR_GRAY),
      type(type)
{
    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new UIButton(_("Close"), SDLK_ESCAPE)));
    g1.add(h.add(new UIButton(_("Go To"), SDLK_RETURN)));
    g1.add(h.add(new UISpacer()));

    add(h.add(new WObjectTypeHeaderTile(type)));
    add(h.add(new WUfoInfoTile(type)));
    add(g1);

    pack();
}

bool
WUfoInfoWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_UP:
         case SDLK_PAGEUP:
         case vk_WheelUp:
            type.setCurrent(type.getPrev(type.getCurrent(), false));
            return true;
         case SDLK_DOWN:
         case SDLK_PAGEDOWN:
         case vk_WheelDown:
            type.setCurrent(type.getNext(type.getCurrent(), false));
            return true;
         case SDLK_RETURN:
            stop(1);
            return true;
         case SDLK_ESCAPE:
            stop(0);
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}



WScreenId
doUfoScreen()
{
    if (getDisplayedTurn().getCurrentUniverse().getUfos().getNumUfos() == 0)
        return WScreenId();

    WUfoInfoWindow win(global_ufo_type);
    ui_root->centerWidget(win);
    int result = ui_root->runModal(win);
    if (result == 1) {
        // "Go to"
        return WScreenId(global_ufo_type.getUfo().getPos());
    } else {
        // "Close"
        return WScreenId();
    }
}

/** Find nearest Ufo. Returns 0 if none found, otherwise slot for GUfoType. */
int
findNearestUfo(GPoint pt)
{
    // FIXME: this has implicit knowledge of the Id assignment scheme of GUfoType
    int found = 0;
    long dist2 = 0;
    bool inside = false;
    int slot = 0;
    const GUfoContainer& ufos = getDisplayedTurn().getCurrentUniverse().getUfos();
    for (GUfoContainer::const_iterator i = ufos.begin(); i != ufos.end(); ++i) {
        long ndist2 = distanceSquared(i->getPos(), pt);
        bool ninside = (ndist2 <= square(i->getRadius()));
        ++slot;
        /* Pick this Ufo, if
           - it's the first one
           - it's closer than our previous choice
           - we're inside it but not inside the previous one */
        if (found <= 0
            || ndist2 < dist2
            || (ninside && !inside))
        {
            found  = slot;
            dist2  = ndist2;
            inside = ninside;
        }
    }
    return found;
}
