/**
  *  \file game/msg-conf.cc
  *  \brief Message Configuration
  *
  *  This module contains the global message configuration, and
  *  functions to load and save it to a config file (msgX.ini).
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/msg.h"
#include "game/gamedirs.h"
#include "io/streams.h"
#include "util/confpars.h"
#include "cpluslib/format.h"

/** Name of the per-player message configuration file. */
static const char MSG_CONFIG_TEMPLATE[] = "msg%d.ini";

/** Global message filter instance.
    \see getMessageFilter() */
static GMessageFilter message_filter;

/** Initialize Message Configuration before start. */
void
initMessageConfig()
{
    message_filter.clear();
}

/** Shut down Message Configuration after termination. */
void
doneMessageConfig()
{
    initMessageConfig();
}

/** Load message configuration for a player. */
void
loadMessageConfig(const int player)
{
    class MessageConfigReader : public ConfigParser {
     public:
        void assign(int /*line_id*/, string_t name, string_t value, string_t /*line*/)
            {
                if (strCaseCmp(name, "filter") == 0)
                    message_filter.addToFilter(value);
            }
    };

    MessageConfigReader mcr;
    mcr.configure("", true);
    mcr.parseGameFile(format(MSG_CONFIG_TEMPLATE, player));
}

/** Save message configuration for a player. */
void
saveMessageConfig(int player)
{
    class MessageConfigWriter : public ConfigParser {
     public:
        bool had_any_content;
        string_t output;

        MessageConfigWriter()
            : had_any_content(false),
              output()
            { }

        void assign(int line_id, string_t name, string_t /*value*/, string_t line)
            {
                if (strCaseCmp(name, "filter") != 0)
                    unparsed(line_id, line);
            }

        void unparsed(int /*line_id*/, string_t line)
            {
                /* We want to erase the file if it has no content. It will, however,
                   always contain our comment line. Hence, we define the file to have
                   "any content" if it contains more than a comment in line 1, not
                   counting blank lines. Thus, we'll also preserve if users add more
                   comments to the file. */
                if (!line.empty()) {
                    if (output.empty() && line[0] == '#') {
                        /* It's the first-line comment */
                    } else {
                        had_any_content = true;
                    }
                }
                output += line;
                output += Platform::LINE_SEPARATOR;
            }
    };

    const string_t filename = format(MSG_CONFIG_TEMPLATE, player);

    // Read existing file
    MessageConfigWriter mcw;
    mcw.configure("", true);
    mcw.parseGameFile(filename);

    // Add filter entries
    for (GMessageFilter::iterator i = message_filter.begin(); i != message_filter.end(); ++i) {
        if (mcw.output.empty()) {
            mcw.output += "# PCC2 Message Configuration File";
            mcw.output += Platform::LINE_SEPARATOR;
        }
        mcw.output += "Filter=";
        mcw.output += *i;
        mcw.output += Platform::LINE_SEPARATOR;
        mcw.had_any_content = true;
    }

    // Create file
    if (!mcw.had_any_content) {
        // Delete it
        game_file_dir->eraseNT(filename);
    } else {
        // Make it
        Ptr<Stream> s = game_file_dir->openFile(filename, Stream::C_CREATE);
        s->writeT(mcw.output.data(), mcw.output.size());
    }
}

/** Get global message filter object. */
GMessageFilter&
getMessageFilter()
{
    return message_filter;
}
