/**
  *  \file vcr.h
  *  \brief VCR Player Interface
  *
  *  The abstract interface for playing battles. This defines the
  *  interface used by players and visualizers.
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_VCR_H
#define PCC_V2_VCR_H

#include "defs.h"
#include "game/struct.h"
#include "cpluslib/smartptr.h"
#include "cpluslib/string.h"

/* This is an optimisation opportunity. It may or may not be faster
   this way, depending on the architecture.
   - pro: on x86, VCR_STATIC saves a register otherwise occupied by
     the frame pointer.
   - con: on RISC, VCR_STATIC may require additional registers to
     generate the addresses, depending heavily on compiler smartness.
   On my Pentium 200 as well as my Athlon XP 2800+, it is faster
   by about 10% when enabled. */
#ifdef CONF_STATIC_VCR
#  define VCR_STATIC static
#else
#  define VCR_STATIC
#endif

class VcrStatus;
class VcrVisualizer;

/// Unpacked VCR Structure.
struct TVcr : public TRawVcr {
    int16 num_fighters[2];
    int16 num_torps[2];
    int16 torp_launcher_count[2];

    TVcr() { }
    TVcr(const TRawVcr& raw);
    bool pack(TRawVcr& raw);
 private:
    bool packSide(TRawVcr& raw, int side);
    void unpackSide(int side);
};

/// Statistics item.
struct VcrStatItem {
    int min_fighters_aboard;
    int torps_hit;
    int num_fights;
};

/// Abstract VCR Player.
class VcrPlayer {
 public:
    typedef long time_type;
 protected:
    VCR_STATIC Ptr<VcrVisualizer> vis;
    VCR_STATIC time_type time;
    VCR_STATIC int status_word;
    VCR_STATIC VcrStatItem stat_info[2];

    void initStat();
    VCR_STATIC void setStatFighters(int side, int count);
    VCR_STATIC void setStatTorpHit(int side);
 public:
    /** Status values (battle outcome). */
    enum {
        VCRS_LEFT_DESTROYED  = 1,  ///< Left object has been destroyed
        VCRS_RIGHT_DESTROYED = 2,  ///< Right object has been destroyed
        VCRS_LEFT_CAPTURED   = 4,  ///< Left object has been captured
        VCRS_RIGHT_CAPTURED  = 8,  ///< Right object has been captured
        VCRS_TIMEOUT         = 16, ///< Battle timed out
        VCRS_STALEMATE       = 32, ///< Stalemate (neither has ammo).
        
        VCRS_INVALID         = -1  ///< Invalid battle
    };
    /** Fighter statuses. Those are hardcoded at many places, don't change! */
    enum {
        FIGHTER_IDLE,
        FIGHTER_ATTACKS,
        FIGHTER_RETURNS
    };
    /** Manifest constants. Id values in the corresponding accessors are
        [0, VCRMF_xxx). */
    enum {
        VCRMF_MAX_BEAMS = 20,   ///< Max number of beams.
        VCRMF_MAX_TORPS = 20,   ///< Max number of torpedo launchers.
        VCRMF_MAX_FTRS  = 50,   ///< Max number of concurrently launched fighters.
        VCRMF_MAX_BAYS  = 50    ///< Max number of fighter bays.
    };
    
    VcrPlayer();
    virtual ~VcrPlayer();

    /** Check VCR. Limits values in \c vcr to those this player can
        handle. Returns true if \c vcr was modified (if this is a
        host-generated VCR, this means it is bogus and should not
        be played). */
    virtual bool checkVcr(TVcr& vcr) = 0;

    /** Init VCR. Initializes the battle to start at time zero. Must
        call initStat(). */
    virtual void initVcr(TVcr& vcr) = 0;

    /** Finish up VCR. Copies status back to the specified VCR
        structure, does explosions and such, and computes the
        final status. */
    virtual void doneVcr(TVcr& vcr) = 0;

    /** Set capabilities. Returns false if capability set not
        supported. */
    virtual bool setCapabilities(uint16 cap) = 0;

    /** Play one cycle. Either does nothing and returns false (last
        cycle), or advances time, does something and returns true. */
    virtual bool playCycle() = 0;

    virtual void playFastForward();

    void playVcr(TVcr& vcr);

    /** \name Accessor Interface */
    //@{
    /** Get beam status.
        \param side 0 or 1
        \param id beam number, starting at 0
        \return value in range [0, 100] (uncharged .. fully charged) */
    virtual int getBeamStatus(int side, int id) = 0;
    /** Get torpedo status.
        \param side 0 or 1
        \param id torp number, starting at 0
        \return value in range [0, 100] (uncharged .. fully charged) */
    virtual int getTorpStatus(int side, int id) = 0;
    /** Get torpedo count (number of torps on ship).
        \param side 0 or 1
        \return number of torpedoes */
    virtual int getTorpCount(int side) = 0;
    /** Get fighter count (number of fighters on ship).
        \param side 0 or 1
        \return number of fighters */
    virtual int getFighterCount(int side) = 0;
    /** Get shield status.
        \param side 0 or 1
        \return value in range [0, 100] (no shields .. full shields, as
        in VGAP rules docs) */
    virtual int getShields(int side) = 0;
    /** Get damage level.
        \param side 0 or 1
        \return damage level. Suggested range is [0, 9999]. VGAP rules
        allow damage levels up to 99 or 150, but allowing 9999 makes a
        Cube-vs.-NFC fight look much more impressive. */
    virtual int getDamage(int side) = 0;
    /** Get crew.
        \param side 0 or 1
        \return number of crewmen */
    virtual int getCrew(int side) = 0;

    /** Get Fighter Position. Positions are [0, 640] */
    virtual int getFighterX(int side, int id) = 0;
    /** Get Fighter Status. One of the FIGHTER_xxx values. */
    virtual int getFighterStatus(int side, int id) = 0;
    /** Get Object Position. Positions are [0, 640] */
    virtual int getObjectX(int side) = 0;
    /** Get distance in meters. */
    virtual long getDistance() = 0;

    /** Get timer. */
    time_type getTime() const
        { return time; }
    /** Get status word. */
    int getStatusWord() const
        { return status_word; }

    /** Get name of this VCR player. */
    virtual string_t getPlayerName() const = 0;
    //@}

    /** Get a status token. The status token can later be passed
        back to setStatus() to restart playing at that place. */
    virtual Ptr<VcrStatus> getStatusToken() = 0;
    /** Set status. status is a status token as returned by
        getStatusToken(). */
    virtual void setStatus(Ptr<VcrStatus> status) = 0;

    void setVisualizer(Ptr<VcrVisualizer> avis);
    Ptr<VcrVisualizer> getVisualizer() const;

    /** Format result as string. */
    static string_t getResultString(int status, const TVcr& vcr);

    /** Get statistical values. */
    const VcrStatItem& getStat(int side) const
        { return stat_info[side]; }

    /** Set statistical values. */
    void setStat(int side, const VcrStatItem& it)
        { stat_info[side] = it; }
};

/// Abstract VCR Status token.
class VcrStatus {
 public:
    VcrStatus(VcrPlayer::time_type atime);
    virtual ~VcrStatus();
    VcrPlayer::time_type time;
};

/// Abstract VCR Visualizer.
class VcrVisualizer {
 protected:
    VcrPlayer* ply;
 public:
    VcrVisualizer();
    virtual ~VcrVisualizer();

    virtual void startFighter(int side, int track) = 0;
    virtual void landFighter(int side, int track) = 0;
    virtual void killFighter(int side, int track) = 0;

    /* hit >= 0: hit, hit < 0: miss */
    /* track >= 0: fighter fires; < 0: beam fires; track = -1-beam nr. */
    virtual void fireBeam(int side, int track, int target, int hit, int damage, int kill) = 0;
    virtual void fireTorp(int side, int hit, int launcher) = 0;

    virtual void updateBeam(int side, int id) = 0;
    virtual void updateTorp(int side, int id) = 0;

    virtual void killObject(int side) = 0;

    virtual void redraw() = 0;

    virtual void init() = 0;

    void setPeer(VcrPlayer* ply);
    VcrPlayer* getPeer() const
        { return ply; }
};

class VcrNullVisualizer : public VcrVisualizer {
 public:
    void startFighter(int side, int track);
    void landFighter(int side, int track);
    void killFighter(int side, int track);
    
    void fireBeam(int side, int track, int target, int hit, int damage, int kill);
    void fireTorp(int side, int hit, int launcher);

    void updateBeam(int side, int id);
    void updateTorp(int side, int id);
    
    void killObject(int side);

    void redraw();

    void init();
};

inline Ptr<VcrVisualizer>
VcrPlayer::getVisualizer() const
{
    return vis;
}

/** Statistics: set number of fighters. Notify the statistics engine
    that \c side now has \c count fighters. */
inline void
VcrPlayer::setStatFighters(int side, int count)
{
    if (stat_info[side].min_fighters_aboard > count)
        stat_info[side].min_fighters_aboard = count;
}

/** Statistics: register that a torpedo has hit. Notify the statistics
    engine that a torpedo fired by \c side has hit its target. */
inline void
VcrPlayer::setStatTorpHit(int side)
{
    ++stat_info[side].torps_hit;
}

#endif
