/**
  *  \file gfx/gfx.h
  *  \brief Graphics Kernel Interface
  *
  *  Prototypes and declarations for graphics kernel (higher-level
  *  API wrapped around SDL).
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_GFX_H
#define PCC_V2_GFX_GFX_H

#include "gfx/gfxtypes.h"
#include "gfx/gfxutil.h"
#include "cpluslib/object.h"

class GfxFont;

/// Graphics Parameter Collection.
struct GfxParameters {
    uint32          color;              ///< Color in target format
    uint32          bgcolor;            ///< Background color, if used
    unsigned        line_thickness;     ///< Line thickness
    uint8           line_style;         ///< Line pattern
    GfxFillPattern  fill_pattern;       ///< Bar fill pattern
    alpha_type      alpha;              ///< Alpha channel
    int             x, y;               ///< Graphics Cursor
    int             just_x, just_y;     ///< Alignment for ::outText()
    GfxFont*        font;               ///< Font for ::outText()

    GfxParameters();

    /** Move graphics cursor to specified position. */
    GfxParameters& moveTo(int ax, int ay)
        { x = ax, y = ay; return *this; }
    /** Move graphics cursor, relative movement. */
    GfxParameters& moveRel(int ax, int ay)
        { x += ax, y += ay; return *this; }
    /** Set foreground color. */
    GfxParameters& setColor(uint32 n)
        { color = n; return *this; }
    /** Set background color. */
    GfxParameters& setBackgroundColor(uint32 n)
        { bgcolor = n; return *this; }
    /** Set line thickness. */
    GfxParameters& setLineThickness(unsigned n)
        { line_thickness = n; return *this; }
    /** Set line style (pattern). */
    GfxParameters& setLineStyle(uint8 n)
        { line_style = n; return *this; }
};

class GfxCanvas : public Object {
 public:
    virtual ~GfxCanvas();

    /** Draw horizontal line.
        \param x1,y1 starting position.
        \param x2    ending X coordinate, not inclusive.
        \param color color, in target format.
        \param pat   pattern. Only draw pixels set in this pattern.
        \param alpha alpha value for transparency. */
    virtual void drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha) = 0;
    /** Draw vertical line.
        \param x1,y1 starting position.
        \param y2    ending Y coordinate, not inclusive.
        \param color color, in target format.
        \param pat   pattern. Only draw pixels set in this pattern.
        \param alpha alpha value for transparency. */
    virtual void drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha) = 0;
    /** Draw single pixel.
        \param x,y   position.
        \param color color, in target format.
        \param alpha alpha value for transparency. */
    virtual void drawPixel(int x, int y, uint32 color, alpha_type alpha) = 0;
    /** Draw filled rectangle.
        \param rect  area
        \param color color, in target format.
        \param bg    background color, or GFX_TRANSPARENT.
        \param pat   fill pattern.
        \param alpha alpha value for transparency. */
    virtual void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha) = 0;
    /** Copy other surface (pixmap).
        \param x,y   anchor point. src's (0,0) is here.
        \param src   input data
        \param rect  display this part of src, relative coordinates.

        `blitSurface(x, y, src, GfxRect(src))' will display the whole
        surface. `blitSurface(x, y, src, GfxRect(2, 3, 4, 5))' will
        display a 4x5 area from surface src, at (x+2, y+3) up to but not
        including (x+2+4, y+3+5). */
    virtual void blitSurface(int x, int y, GfxCanvas& src, GfxRect rect) = 0;
    /** Display pattern (monochrome pixmap).
        \param rect    position, screen coordinates;
        \param x,y     position of bit 7, data[0], on the screen;
        \param bytes_per_line  bytes per line in data;
        \param data    pixel data. bit 7 = left, bit 0 = right (VGA bit order);
        \param color   color of "on" bits;
        \param bg      color of "off" bits, or GFX_TRANSPARENT;
        \param alpha   alpha channel.

        \pre x <= rect.x, y <= rect.y

        Note that this function has no constraints on the bytes_per_line
        parameter. bytes_per_line can also be zero or negative for
        interesting effects. */
    virtual void blitPattern(GfxRect rect, int x, int y, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha) = 0;
    /** Lock surface. Users must call this before any output
        operation. Recursive locks are permitted. */
    virtual bool lock() = 0;
    /** Unlock surface. Users must call this after any output
        operation. */
    virtual void unlock() = 0;
    /** Update part of output. This will make sure that changes inside
        r will be shown on the screen. Without update, output might or
        might not happen, depending on the type of surface. */
    virtual void update(GfxRect r) = 0;
    /** Compute clipping rectangle. This can be used to optimize
        output. This computes the smallest rectangle p such that
        drawBar(p) and drawBar(r) have the same effect. */
    virtual GfxRect computeClipRect(GfxRect r) = 0;
    /** Read one pixel. \return pixel value in target format. */
    virtual uint32 getPixel(int x1, int y1) = 0;
    /** Get size of this surface */
    virtual GfxPoint getSize() = 0;
    /** Get number of bits used per pixel. */
    virtual int getBitsPerPixel() = 0;

    virtual bool isVisible(GfxRect r);
    virtual bool isClipped(GfxRect r);
};

class GfxSurface : public GfxCanvas {
 protected:
    SDL_Surface* sfc;
    bool owner;
    int locked;

    void setSurface(SDL_Surface* asfc, bool aown = true);

 public:
    GfxSurface(GfxSystemSurface asfc, bool aowner);
    GfxSurface(int wi, int he, GfxSystemPixelFormat fmt);
    ~GfxSurface();

    void drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha);
    void drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha);
    void drawPixel(int x, int y, uint32 color, alpha_type alpha);
    void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha);

    void blitSurface(int x, int y, GfxCanvas& src, GfxRect rect);
    void blitPattern(GfxRect rect, int x, int y, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha);

    bool lock();
    void unlock();
    void update(GfxRect r);

    GfxRect computeClipRect(GfxRect r);
    GfxPoint getSize();
    int getBitsPerPixel();

    GfxSystemSurface getSystemSurface();

    uint32 getPixel(int x, int y);
};

class GfxFilter : public GfxCanvas {
 protected:
    GfxCanvas& parent;
    GfxFilter(GfxCanvas& parent);
 public:
    GfxCanvas& getParent() const { return parent; }

    bool lock();
    void unlock();
    void update(GfxRect r);

    GfxPoint getSize();
    int getBitsPerPixel();

    uint32 getPixel(int x, int y);
};

class GfxXorFilter : public GfxCanvas {
    GfxCanvas& parent;
 public:
    GfxXorFilter(GfxCanvas& parent);
    ~GfxXorFilter();

    void drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha);
    void drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha);
    void drawPixel(int x, int y, uint32 color, alpha_type alpha);
    void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha);

    void blitSurface(int x, int y, GfxCanvas& src, GfxRect rect);
    void blitPattern(GfxRect rect, int x, int y, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha);

    bool lock();
    void unlock();
    void update(GfxRect r);

    GfxRect computeClipRect(GfxRect r);
    GfxPoint getSize();
    int getBitsPerPixel();

    uint32 getPixel(int x, int y);
};

class GfxNullCanvas : public GfxCanvas {
 public:
    GfxNullCanvas() { }
    ~GfxNullCanvas();

    void drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha);
    void drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha);
    void drawPixel(int x, int y, uint32 color, alpha_type alpha);
    void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha);

    void blitSurface(int x, int y, GfxCanvas& src, GfxRect rect);
    void blitPattern(GfxRect rect, int x, int y, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha);

    bool lock();
    void unlock();
    void update(GfxRect r);

    GfxRect computeClipRect(GfxRect r);
    GfxPoint getSize();
    int getBitsPerPixel();

    uint32 getPixel(int x, int y);
};

// re-export these functions:
#include "gfx/gfxcompl.h"

#endif
