#!/usr/bin/perl -w
#
#  This script builds a directory tree containing the Windows binary
#  installation of PCC2. It can operate in three modes. The mode is
#  specified as first parameter to this script, as in
#  "perl prepare2.pl list".
#
#  list [default]: Just display the files that would be installed
#     on standard output.
#
#  script: Generate a ".nsh" file on standard output. The ".nsh" file
#     will contain macros for installation and uninstallation, and can
#     be referred to from the main ".nsi" file.
#
#  install: Takes a directory name as second parameter. This will
#     perform an installation.
#
#  To run this script on your computer, please configure the path names
#  below first. As is, this script assumes SDL_image-1.2.1 and SDL-1.2.2
#  sitting next to the "pcc-v2" folder in your file system.
#

use strict;

$| = 1;

my $mode = shift(@ARGV) || 'list';
my $return = 0;

# Path name configuration
my $basedir = "../..";
my $srcdir  = "..";
my $SDL_image = "$basedir/SDL_image-1.2.1/lib";
my $SDL       = "$basedir/SDL-1.2.2/lib";

# $stuffToInstall{Component} = [ [source => dest], [source => dest], ... ];
my %stuffToInstall;

# Step 1: DLLs
addFile('DLLs', "$SDL/SDL.dll");
addFile('DLLs', "$SDL_image/SDL_image.dll");
addFile('DLLs', "$SDL_image/jpeg.dll");
addFile('DLLs', "$SDL_image/libpng1.dll");
addFile('DLLs', "$SDL_image/zlib.dll");

# Step 2: Base files
addDirectory('BaseFiles', "$srcdir/specs", "specs");
addDirectory('BaseFiles', "$srcdir/resource", "resource", undef, "\\.lang\$");
addDirectory('Language', "$srcdir/resource", "resource", "\\.lang\$", undef);

# Some documentation files. Note that naming them ".doc" causes them to
# be opened with Wordpad by default, which, unlike Notepad, supports
# Unix-style line endings
addFile('BaseFiles', "$srcdir/README", "doc/README.doc");
addFile('BaseFiles', "$srcdir/NEWS", "doc/NEWS.doc");
addFile('BaseFiles', "$srcdir/AUTHORS", "doc/AUTHORS.doc");
addFile('BaseFiles', "$srcdir/COPYING", "doc/COPYING.doc");
addFile('BaseFiles', "$srcdir/doc/pcc2-overview.html", "doc/pcc2-overview.html");
addFile('PlayVCR', "$srcdir/doc/playvcr.txt", "doc/playvcr.doc");

# Step 3: Client
addFile('Client', "$srcdir/pcc-v2.exe");
addFile('PlayVCR', "$srcdir/playvcr.exe");
addFile('UnTRN', "$srcdir/un-trn.exe");
addDirectory('CommandLine', "$srcdir", ".", "c2.*\\.exe\$", undef);

addFile('MapFiles', "$srcdir/pcc-v2.map");
addFile('MapFiles', "$srcdir/playvcr.map");
addFile('MapFiles', "$srcdir/un-trn.map");

# Step 4: Time stamp
my @time = localtime();
open TIMESTAMP, "> version.txt";
printf TIMESTAMP "This package was built on %04d%02d%02d at %02d:%02d:%02d\n",
    $time[5] + 1900, $time[4] + 1, $time[3], $time[2], $time[1], $time[0];
close TIMESTAMP;
addFile('BaseFiles', 'version.txt');

open TIMESTAMP, "> version.nsh";
printf TIMESTAMP "!define TIMESTAMP \"%04d%02d%02d\"\n", $time[5] + 1900, $time[4] + 1, $time[3];
printf TIMESTAMP "!define TIMESTAMP_YEAR \"%d\"\n", $time[5] + 1900;
printf TIMESTAMP "!define TIMESTAMP_MONTH \"%d\"\n", $time[4] + 1;
printf TIMESTAMP "!define TIMESTAMP_DAY \"%d\"\n", $time[3];
close TIMESTAMP;

# Final step: Generate Output
if ($mode eq 'list') {
    # List in human-readable form
    foreach (sort keys %stuffToInstall) {
        print "$_:\n";
        foreach (@{$stuffToInstall{$_}}) {
            print "  $_->[0] => $_->[1]\n";
        }
    }
} elsif ($mode eq 'script') {
    # Generate installer functions
    my %seenDirectories;
    foreach my $section (sort keys %stuffToInstall) {
        my %seenDirectoriesInThisSection;
        foreach (@{$stuffToInstall{$section}}) {
            $_->[1] =~ m|(.*)/| and $seenDirectories{$1} = $seenDirectoriesInThisSection{$1} = 1;
        }

        print "!macro Install$section\n";
        foreach (sort {length($a)<=>length($b)} keys %seenDirectoriesInThisSection) {
            print "  CreateDirectory \"\$INSTDIR\\" . flipSlash($_) . "\"\n";
        }
        foreach (@{$stuffToInstall{$section}}) {
            print "  File \"/oname=" . flipSlash($_->[1]) ."\" \"" . flipSlash($_->[0]) . "\"\n";
        }
        print "!macroend\n\n";

        print "!macro un.Install$section\n";
        foreach (@{$stuffToInstall{$section}}) {
            print "  Delete \"\$INSTDIR\\" . flipSlash($_->[1]) . "\"\n";
        }
        print "!macroend\n\n";
    }
    print "!macro un.RemoveDirs\n";
    foreach (sort {length($b)<=>length($a)} keys %seenDirectories) {
        print "  RMDir \"\$INSTDIR\\" . flipSlash($_) . "\"\n";
    }
    print "!macroend\n\n";
    print "!macro un.InstallAll\n";
    foreach (sort keys %stuffToInstall) {
        print "  !insertmacro un.Install$_\n";
    }
    print "!macroend\n\n";
} elsif ($mode eq 'install') {
    # Manual installation
    my $installdir = shift @ARGV;
    if (!defined($installdir)) {
        print STDERR "Please specify target directory.\n";
        $return = 1;
    } else {
        foreach my $section (sort keys %stuffToInstall) {
            foreach (@{$stuffToInstall{$section}}) {
                installFile($_->[0], "$installdir/$_->[1]");
            }
        }
    }
} else {
    print STDERR "Unknown mode '$mode': valid modes are:\n";
    print STDERR "    list          List in human-readable format\n";
    print STDERR "    script        Write NSIS macro script to stdout\n";
    print STDERR "    install DIR   Install to DIR\n";
}

exit $return;


sub addDirectory {
    my ($group, $source, $dest, $includeRX, $excludeRX) = @_;
    my @todo = ([$source, $dest]);
    my $filesAdded = 0;
    while (@todo) {
        my $item = shift @todo;
        if (opendir DIR, $item->[0]) {
            while (defined(my $entry = readdir(DIR))) {
                if (uc($entry) eq 'CVS' || $entry =~ /~$/ || $entry =~ /^\./ || $entry =~ /\.bak$/i) {
                    # ignore
                } elsif (defined($excludeRX) && $entry =~ /$excludeRX/) {
                    # ignore, because excludeRX given and file does match
                } elsif (defined($includeRX) && $entry !~ /$includeRX/) {
                    # ignore, because includeRX given but file does not match
                } elsif (-d "$item->[0]/$entry") {
                    # Subdirectory
                    push @todo, ["$item->[0]/$entry", "$item->[1]/$entry"];
                } else {
                    # File
                    addFile($group, "$item->[0]/$entry", "$item->[1]/$entry");
                    ++$filesAdded;
                }
            }
            closedir DIR;
        } else {
            print STDERR "Unable to read $item->[0]: $!\n";
            $return = 1;
        }
    }
    if ($filesAdded == 0) {
        print STDERR "Scan for '$group' in '$source' didn't match any files\n";
        $return = 1;
    }
}

sub addFile {
    my ($group, $source, $dest) = @_;
    my $basename = $source;

    if (!-r $source) {
        print STDERR "File $source does not exist\n";
        $return = 1;
    }

    $basename =~ s|^.*/||;
    if (!defined($dest)) {
        $dest = $basename;
    } elsif ($dest =~ m|/$|) {
        $dest .= $basename;
    }

    $dest =~ s|^\./||;
    $dest =~ s|^/\./|/|;

    push @{$stuffToInstall{$group}}, [ $source => $dest ];
}

sub flipSlash {
    my $x = shift;
    $x =~ s|/|\\|g;
    $x;
}

sub installFile {
    my $src = shift;
    my $dest = shift;
    createDirectoryFor($dest);
    run("copy /Y \"" . flipSlash($src) . "\" \"" . flipSlash($dest) . "\"");
}

sub createDirectoryFor {
    my $fn = shift;
    if ($fn =~ m|^(.*)/|) {
        my $dir = $1;
        if (!-d $dir) {
            createDirectoryFor($dir);
            run("mkdir \"" . flipSlash($dir) . "\"");
        }
    }
}

sub run {
    my $command = shift;
    print "\$ $command\n";
    if (system($command) != 0) {
        print STDERR "Command failed!\n";
        $return = 1;
    }
}
