/**
  *  \file util/atom.cc
  *  \brief Atom Table
  *
  *  This module manages atoms. An atom is an integer representing a
  *  string. A fast mapping between strings and atoms is provided.
  *
  *  Differences to the PCC 1.x implementation:
  *  - we have no real upper bound for atoms.
  *  - there are functions atomNC to inquire the atom table without
  *    creating an atom. These should not be used in user code except
  *    as an optimisation.
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <utility>
#include <vector>
#include "util/atom.h"
#include "util/misc.h"
#include "cpluslib/misc.h"
#include "cpluslib/assert.h"

namespace {
    /** Manifest constants. */
    enum {
        /** Atoms per memory chunk. */
        ATOMS_PER_CHUNK = 128,
        /** Number of hash-table slots. */
        ATOM_HASH       = 128
    };

    struct AtomChunk {
        /** All atoms of this chunk in one block */
        string_t data;
        /** Positions of each atom */
        string_t::size_type pos[ATOMS_PER_CHUNK];
        string_t::size_type length[ATOMS_PER_CHUNK];
        /** Hash links for each atom */
        atom_t   hash_next[ATOMS_PER_CHUNK];
    };

    struct AtomTable {
        std::vector<AtomChunk*> entry_vec;
        atom_t hash_first[ATOM_HASH];
        atom_t next;

        AtomTable();
        ~AtomTable();
    };

    typedef std::pair<const char*, string_t::size_type> AtomDef;

    static AtomTable tab;
};

AtomTable::AtomTable()
{
    std::fill_n(hash_first, countof(hash_first), 0);
    next = ATOM_LOWER_BOUND;
}

AtomTable::~AtomTable()
{
    safeClear(entry_vec);
}

/** Look up definition of an atom. Returns address/length pair.
    \pre isAtom(atom) */
static AtomDef
atomStrLookup(atom_t atom)
{
    ASSERT(atom >= ATOM_LOWER_BOUND && atom < tab.next);
    atom -= ATOM_LOWER_BOUND;
    unsigned slot = atom % ATOMS_PER_CHUNK;
    unsigned blk  = atom / ATOMS_PER_CHUNK;
    const AtomChunk& pac = *tab.entry_vec[blk];
    return std::make_pair(pac.data.data() + pac.pos[slot],
                          pac.length[slot]);
}

/** Look up value of an atom. This is the common back-end of atom() and
    atomNC().
    \param data    text
    \param length  its length
    \param add     when true, this creates the atom when its not found
    \return atom if found or newly-created, NULL_ATOM if not found and
    add was false. */
static atom_t
atomLookup(const char* data, std::size_t length, bool add)
{
    if (!length)
        return NULL_ATOM;

    unsigned hv = strHash(data, length) % ATOM_HASH;
    atom_t   index = tab.hash_first[hv];
    while (index) {
        AtomDef ad = atomStrLookup(index);
        if (ad.second == length && std::memcmp(ad.first, data, length) == 0)
            return index;
        index -= ATOM_LOWER_BOUND;
        index = tab.entry_vec[index / ATOMS_PER_CHUNK]->hash_next[index % ATOMS_PER_CHUNK];
    }
    if (!add)
        return NULL_ATOM;

    index = tab.next - ATOM_LOWER_BOUND;
    unsigned slot = index % ATOMS_PER_CHUNK;
    unsigned blk  = index / ATOMS_PER_CHUNK;
    if (slot == 0) {
        /* we're starting a new block */
        ASSERT(tab.entry_vec.size() == blk);
        safeAppend(tab.entry_vec, new AtomChunk());
    }
    AtomChunk& pac = *tab.entry_vec[blk];
    pac.pos[slot] = pac.data.length();
    pac.length[slot] = length;
    pac.data.append(data, length);
    pac.hash_next[slot] = tab.hash_first[hv];
    tab.hash_first[hv] = tab.next;
    return tab.next++;
}

/** Get atom corresponding to string str.
    \param str Atom to look up / create
    \return atom number
    \post str!="" => isAtom(return) && atomStr(return)==str */
atom_t
atom(const char* str)
{
    return atomLookup(str, std::strlen(str), true);
}

/** Get atom corresponding to string str.
    \overload */
atom_t
atom(const string_t& str)
{
    return atomLookup(str.data(), str.length(), true);
}

/** Look up atom corresponding to string str, but does not create it if
    there isn't one already.
    \param str [in] C-string
    \return atom, 0 if not found or str=="" */
atom_t
atomNC(const char* str)
{
    return atomLookup(str, std::strlen(str), false);
}

/** Look up atom corresponding to string str, but does not create it if
    there isn't one already. \overload */
atom_t
atomNC(const string_t& str)
{
    return atomLookup(str.data(), str.length(), false);
}

/** Return string associated with an atom. Returns empty string if atom
    is invalid. */
string_t
atomStr(atom_t atom)
{
    if (atom < ATOM_LOWER_BOUND || atom >= tab.next) {
        return string_t();
    } else {
        AtomDef d = atomStrLookup(atom);
        return string_t(d.first, d.second);
    }
}

/** True iff atom corresponds to a string. */
bool
isAtom(atom_t atom)
{
    return (atom >= ATOM_LOWER_BOUND && atom < tab.next);
}

/** Compute hash value over a character array. For now, we simply add
    all characters.
    \param str    [in] Array
    \param length [in] Size
    \return hash value */
unsigned int
strHash(const char* str, std::size_t length)
{
    unsigned int sum = 0;
    while (length--)
        sum += static_cast<unsigned char>(*str++);
    return sum;
}

/** Compute hash value over a C string. \overload */
unsigned int
strHash(const char* str)
{
    return strHash(str, std::strlen(str));
}
