/**
  *  \file confpars.cc
  *  \brief Configuration Parser
  *
  *  Logic for parsing pconfig.src-like config files.
  *
  *  (c) 2001-2009 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <istream>
#include "util/confpars.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/string.h"

/** \class ConfigParser
    \brief Configuration File Parser (pconfig.src alike)

    This class encapsulates the logic for parsing PCONFIG.SRC-alike
    configuration files. In a nutshell:
    - sections separated by `% sectionname'
    - comments starting with `\#'
    - assignments `key = value' in each section

    To parse a configuration file, derive a class from this and implement
    at least the assign() method. */

ConfigParser::ConfigParser()
    : FileParser("#"),
      section(),
      in_section(true)
{ }

/** Syntax error. This function is called when the parser encounters
    a syntax error.

    \param line_id line number
    \param s       error message

    \default Throws a FileFormatException. */
void
ConfigParser::error(int line_id, string_t s)
{
    throw FileFormatException(getFileName(), itoa(line_id) + ": " + s);
}

void
ConfigParser::configure(string_t section, bool in_section)
{
    this->section    = section;
    this->in_section = in_section;
}

bool
ConfigParser::isInSection() const
{
    return in_section;
}

string_t
ConfigParser::getSectionName() const
{
    return section;
}

void
ConfigParser::process(int line_id, string_t line)
{
    // Remove whitespace
    string_t ppline = strTrim(line);

    // blank line?
    if (ppline.empty()) {
        unparsed(line_id, line);
        return;
    }

    // Section delimiter?
    if (ppline[0] == '%') {
        if (strCaseCmp(section, strLTrim(ppline.substr(1))) == 0)
            in_section = true;
        else
            in_section = false;
        unparsed(line_id, line);
        return;
    }

    // Process this line further?
    if (!in_section) {
        unparsed(line_id, line);
        return;
    }

    // It's an assignment
    string_t::size_type eqpos = ppline.find('=');
    if (eqpos == string_t::npos) {
        error(line_id, _("Syntax error"));
        return;
    }

    string_t key = strRTrim(ppline.substr(0, eqpos));
    if (key.empty()) {
        error(line_id, _("Syntax error"));
        return;
    }

    assign(line_id, key, strLTrim(ppline.substr(eqpos+1)), line);
}
